# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Simple data types without dependencies on other gazoo_device modules."""
import enum

import dataclasses


@enum.unique
class KeyType(enum.Enum):
  SSH = "ssh"
  OTHER = "other"


@dataclasses.dataclass(frozen=True)
class KeyInfo:
  file_name: str  # Name of the key file.
  type: KeyType  # Key type, SSH or other (such as API).
  package: str  # Name of the package registering the key.



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the switch_power_dli_powerswitch capability."""
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import switch_power_base

logger = gdm_logger.get_logger()

OFF = "off"
ON = "on"


class SwitchPowerDliPowerswitch(switch_power_base.SwitchPowerBase):
  """Definition of the switch_power_dli_powerswitch capability."""
  _REQUIRED_COMMANDS = ["ADJUST_PORTS_MODE"]
  _REQUIRED_HEADERS = ["GET_PROP", "SET_PROP"]

  def __init__(self, http_fn, ip_address, command_dict, headers_dict,
               device_name, total_ports):
    """Create an instance of the powerswitch power capability.

    Args:
        http_fn (func): A function to send GET and PUT http commands
        ip_address (str): The IP address of the Powerswitch.
        command_dict (dict): A dictionary containing the command used for
          each method.
                             The dictionary must contain the following keys:
                               - ADJUST_PORTS_MODE
        headers_dict (dict): A dictionary containg the required headers for
          the HTTP requests. - GET_PROP - SET_PROP
        device_name (str): name of the device this capability is attached
          to.
        total_ports (int): number of ports on device
    """

    super().__init__(device_name=device_name)
    self._http_fn = http_fn
    self._ip_address = ip_address
    self._command_dict = command_dict
    self._headers_dict = headers_dict
    self._total_ports = total_ports

    self.validate_required_keys(self._REQUIRED_COMMANDS, command_dict,
                                "command_dict")
    self.validate_required_keys(self._REQUIRED_HEADERS, headers_dict,
                                "headers_dict")

  @decorators.PersistentProperty
  def supported_modes(self):
    """Get the Powerswitch power modes supported by the dli_powerswitch."""
    return [OFF, ON]

  @decorators.PersistentProperty
  def total_ports(self):
    """Get the number of device ports.

    Returns:
        int: the number of device ports
    """
    return int(self._total_ports)

  def get_all_ports_mode(self):
    """Get mode of all the device ports.

    Returns:
        list: a list of the all the ports mode.
    """
    response = self._http_fn(
        "GET",
        self._command_dict["ADJUST_PORTS_MODE"].format(
            "all;", ip=self._ip_address),
        headers=self._headers_dict["GET_PROP"])
    ports_value = response.split(",")
    ports_mode = [ON if port == "true" else OFF for port in ports_value]
    return ports_mode

  def get_mode(self, port):
    """Get mode of the specified port.

    Args:
        port (int): device port number

    Raises:
        DeviceError: if port is None, port < 0, or port >= total_ports

    Returns:
        str: mode of specified port, either 'on' or 'off'.
    """
    self._validate_port("get_mode", port)
    response = self._http_fn(
        "GET",
        self._command_dict["ADJUST_PORTS_MODE"].format(
            "=" + str(port), ip=self._ip_address),
        headers=self._headers_dict["GET_PROP"])
    if response == "true":
      return ON
    else:
      return OFF

  @decorators.CapabilityLogDecorator(logger)
  def power_on(self, port):
    """Powers on the specified port.

    Args:
        port (int): device port number

    Raises:
        DeviceError: if port is None, port < 0, or port >= total_ports
    """
    self._validate_port("power_on", port)
    logger.debug("{} Powering on powerswitch port {}".format(
        self._device_name, port))
    self._http_fn(
        "POST",
        self._command_dict["ADJUST_PORTS_MODE"].format(
            "=" + str(port), ip=self._ip_address),
        headers=self._headers_dict["SET_PROP"],
        data={"value": "true"})

  @decorators.CapabilityLogDecorator(logger)
  def power_off(self, port):
    """Powers off the specified port.

    Args:
        port (int): device port number

    Raises:
        DeviceError: if port is None, port < 0, or port >= total_ports
    """
    self._validate_port("power_off", port)
    logger.debug("{} Powering off powerswitch port {}".format(
        self._device_name, port))
    self._http_fn(
        "POST",
        self._command_dict["ADJUST_PORTS_MODE"].format(
            "=" + str(port), ip=self._ip_address),
        headers=self._headers_dict["SET_PROP"],
        data={"value": "false"})

  @decorators.CapabilityLogDecorator(logger)
  def set_mode(self, mode, port):
    """Sets the given Powerswitch port to the mode specified.

    Args:
        mode (str): mode to set Powerswitch ports to. e.g. 'off', 'on'
        port (int): the port to set.

    Raises:
        DeviceError: invalid mode.
    """
    self._validate_mode(mode)
    if mode == ON:
      self.power_on(port=port)
    else:
      self.power_off(port=port)

  @decorators.CapabilityLogDecorator(logger)
  def set_all_ports_mode(self, mode):
    """Set all device ports to the specified mode.

    Args:
        mode (str): Mode to set all Powerswitch ports to. Valid modes are
          "on" or "off".

    Raises:
        DeviceError: if input mode is not either 'on' or 'off'
    """
    self._validate_mode(mode)
    if mode == ON:
      data_value = "true"
    else:
      data_value = "false"

    self._http_fn(
        "POST",
        self._command_dict["ADJUST_PORTS_MODE"].format(
            "all;", ip=self._ip_address),
        headers=self._headers_dict["SET_PROP"],
        data={"value": data_value})

  def _validate_mode(self, mode):
    """Verify mode given resides in the valid mode list.

    Args:
        mode (str): mode e.g. 'off', 'on'

    Raises:
        DeviceError: if mode given is not found in the valid mode list.
    """
    if mode not in list(self.supported_modes):
      raise errors.DeviceError(
          "Mode {} not found in supported modes {!r}".format(
              mode, self.supported_modes))

  def _validate_port(self, method_name, port):
    """Ensure port is a valid port number.

    Args:
        method_name(str): name of the method where _validate_port is being
          called
        port (int): device port number

    Raises:
         DeviceError: if input port is not valid
    """
    if port is None or port <= -1 or port >= self._total_ports:
      raise errors.DeviceError("Device {} {} failed. "
                               "Port {} is invalid.".format(
                                   self._device_name, method_name, port))



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Communication Power Default Capability."""
import time
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger

from gazoo_device.capabilities.interfaces import comm_power_base
from gazoo_device.capabilities.interfaces import switch_power_base
from gazoo_device.utility import deprecation_utils

logger = gdm_logger.get_logger()

SUPPORTED_HUB_TYPES = ["cambrionix", "ethernet_switch"]


class CommPowerDefault(comm_power_base.CommPowerBase):
  """Default class for comm_power capability."""

  def __init__(self,
               device_name,
               create_device_func,
               hub_type,
               props,
               settable,
               hub_name_prop,
               port_prop,
               get_switchboard_if_initialized,
               power_and_data_share_cable=False,
               wait_for_connection_func=None,
               wait_for_bootup_complete_func=None,
               pre_off_func=None):
    """Create an instance of the comm_power_default capability.

    Args:
      device_name (str): name of the device this capability is attached
        to.
      create_device_func (func): create_device method.
      hub_type (str): type of power hub for power cycling.
      props (dict): dictionary of device props from configuration file.
      settable (bool): whether or not the properties are settable.
      hub_name_prop (str):  name of the hub name property
      port_prop (str): name of the hub port property
      get_switchboard_if_initialized (callable): function which returns
        a Switchboard instance or None if Switchboard hasn't been initialized.
      power_and_data_share_cable (bool): set power_and_data_share_cable to
        TRUE if device shares a common USB cable for its power and data.
      wait_for_connection_func (func): a method to wait for the device to
        become reachable.
      wait_for_bootup_complete_func (func): a method to call if power and
        data share cable.
      pre_off_func (func): a method to call before turning power off.

    Raises:
      ValueError: If hub_type unknown.
    """
    super().__init__(device_name=device_name)

    self._create_device_func = create_device_func
    self._settable = settable
    self._hub_name_prop = hub_name_prop
    self._port_prop = port_prop
    if hub_type not in SUPPORTED_HUB_TYPES:
      raise ValueError("Hub Type {} not supported. Valid types: {}".format(
          hub_type, SUPPORTED_HUB_TYPES))
    self._hub_type = hub_type

    # Set the properties
    self._dict_name = "optional" if self._settable else "persistent_identifiers"
    self._props = props
    self._get_switchboard_if_initialized = get_switchboard_if_initialized
    self._power_and_data_share_cable = power_and_data_share_cable
    self._pre_off_func = pre_off_func
    self._wait_for_connection_func = wait_for_connection_func
    self._wait_for_bootup_complete_func = wait_for_bootup_complete_func
    self._hub = None

  @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
  def health_check(self):
    """Checks that the capability is ready to use.

    Raises:
        CapabilityNotReadyError: if unable to create auxiliary device for
        power switching.
    """
    unset_props = []
    if not self.hub_name:
      unset_props.append(self._hub_name_prop)
    if not self.port_number:
      unset_props.append(self._port_prop)
    if unset_props:
      if self._settable:
        msg_format = ("If device is connected to {}, "
                      "set them via 'gdm set-prop {} <property> <value>'")

      else:
        msg_format = ("If device is connected to {}, "
                      "set them via 'gdm redetect {}")
      msg = msg_format.format(self._hub_type, self._device_name)
      error_msg = "properties {} are unset. ".format(
          " and ".join(unset_props)) + msg
      raise errors.CapabilityNotReadyError(
          msg=error_msg, device_name=self._device_name)

    try:
      self._hub = self._create_device_func(self.hub_name)
    except errors.DeviceError as err:
      raise errors.CapabilityNotReadyError(
          msg=str(err), device_name=self._device_name)
    self._verify_switch_created(self._hub)
    self._healthy = True

  @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
  def close(self):
    """Closes the hub device instance."""
    if hasattr(self._hub, "close"):  # self._hub may not be a device instance.
      self._hub.close()
    super().close()

  @decorators.OptionalProperty
  def hub_name(self):
    """Name of the hub the device is attached to."""
    return self._props[self._dict_name].get(self._hub_name_prop)

  @decorators.PersistentProperty
  def hub_type(self):
    """Device type of the hub."""
    return self._hub_type

  @decorators.DynamicProperty
  def port_mode(self):
    """port mode."""
    if not self.healthy:
      self.health_check()
    return self._hub.switch_power.get_mode(self.port_number)

  @decorators.OptionalProperty
  def port_number(self):
    """Port number the device is attached to."""
    return self._props[self._dict_name].get(self._port_prop)

  @decorators.CapabilityLogDecorator(logger)
  def cycle(self):
    """Power off then power on the communication port."""
    if not self.healthy:
      self.health_check()
    self.off()
    time.sleep(2)  # Small delay before calling power_on
    self.on()

  @decorators.CapabilityLogDecorator(logger)
  def off(self):
    """Turn off power to the device communications port."""
    if not self.healthy:
      self.health_check()
    if self._pre_off_func:
      self._pre_off_func()
    switchboard = self._get_switchboard_if_initialized()
    if switchboard:
      switchboard.close_all_transports()
      if self._power_and_data_share_cable:
        switchboard.add_log_note(
            f"comm_power.off() called on {self._device_name} set communication "
            f"port {self.port_number} to charge as device has a single USB "
            "cable for data and power.")
    if self._power_and_data_share_cable:
      self._hub.switch_power.power_on(self.port_number, data_sync=False)
    else:
      self._hub.switch_power.power_off(self.port_number)

  @decorators.CapabilityLogDecorator(logger)
  def on(self):
    """Turn on power to the device communication port."""
    if not self.healthy:
      self.health_check()
    self._hub.switch_power.power_on(self.port_number)
    # Wait for 'dev/tty' to populate
    if self._wait_for_connection_func:
      self._wait_for_connection_func()
    else:
      time.sleep(5)
    switchboard = self._get_switchboard_if_initialized()
    if switchboard:
      switchboard.open_all_transports()
    if self._wait_for_bootup_complete_func:
      self._wait_for_bootup_complete_func()

  def _verify_switch_created(self, switch):
    """Verifies switch is created and has switch_power capability."""
    if not (
        hasattr(switch, "switch_power") and
        isinstance(switch.switch_power, switch_power_base.SwitchPowerBase)):
      raise errors.CapabilityNotReadyError(
          msg="'switch_power' capability is missing in hub device {} ({}),"
          " or is not an instance of SwitchPowerBase".format(
              self.hub_name,
              type(switch).__name__),
          device_name=self._device_name)


deprecation_utils.add_deprecated_attributes(CommPowerDefault,
                                            [("enable", "on", True),
                                             ("disable", "off", True),
                                             ("power_cycle", "cycle", True),
                                             ("port", "port_number", False),
                                             ("address", "hub_name", False)])



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility module for working with USB mass storage device.

This module uses a combination of ``pyudev`` and ``psutil`` to list and filter
USB mass storage devices.

.. note:

    According to the udiskctl man page
    http://manpages.ubuntu.com/manpages/trusty/man8/udisks.8.html#contenttoc5

    Therefore, applications can not rely on tools like e.g.  udisksctl(1) to be
    available. Additionally, there is no guarantee that the options,
    command-line switches etc. of command-line tools or similar will remain
    stable.

    Instead, applications should only use the D-Bus API, the libudisks2
    library or tools such as dbus-send(1) or gdbus(1) to interact with
    udisksd(8).
"""
import os
import re
import shutil
import stat
import subprocess
import sys
import time
import psutil
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.utility import usb_utils

logger = gdm_logger.get_logger()


def _validate_suid_bit(cmd_str):
  if sys.platform == "darwin":
    return
  binary = cmd_str.split()[0]
  if not os.stat(binary).st_mode and stat.S_ISUID:
    raise errors.DeviceError(
        "Missing SUID bit on {binary} "
        "Fix using 'sudo chmod u+s {binary}'".format(binary=binary))


class UsbStorage(object):
  """USB mass storage device class.

  Used in 'with block' and will mount the USB mass storage device upon entering
  the with block and eject the USB mass storage device upon exiting the
  'with block'. If the UsbStorage class is not used within a 'with block' then
  the methods of mount, unmount, and eject can be called separately.
  """
  commands_dict = {
      "darwin": {
          "EJECT": "diskutil eject {}",
          "MOUNT": "diskutil mount {}",
          "UNMOUNT": "diskutil unmount {}"
      },
      "linux": {
          "EJECT": "/usr/bin/eject {}",
          "MOUNT": "/usr/bin/udisksctl mount --no-user-interaction -b {}",
          "UNMOUNT": "/bin/umount {}"
      }
  }

  def __init__(self, serial_number, add_log_note_func=None, eject_on_exit=True):
    """Initialize the USB mass storage device instance with the information specified.

    Args:
        serial_number (str): serial number to identify device.
        add_log_note_func (func): function to log usb serial methods to
          device logs.
        eject_on_exit (bool): whether to eject or unmount on exit from
          context manager.
    """
    self._serial_number = serial_number
    self.add_log_note_func = add_log_note_func
    self.eject_on_exit = eject_on_exit
    if sys.platform == "darwin":
      self.commands = self.commands_dict["darwin"]
    else:
      self.commands = self.commands_dict["linux"]

  def __enter__(self):
    self.mount()
    return self

  def __exit__(self, exc_type, exc_val, exc_tb):
    if self.eject_on_exit:
      self.eject()
    else:
      self.unmount()

  @property
  def disk(self):
    """Disk that device mounted to. Empty if not mounted."""
    devices_info = self._get_devices_info()
    return devices_info[0].disk if devices_info else ""

  @property
  def mount_point(self):
    """Returns the mass storage mount point to the caller.

    Returns:
        str: mount path location for caller to use in opening files on the
        mass storage device
    """
    usb_endpoint_infos = self._get_devices_info()
    if usb_endpoint_infos:
      partitions = psutil.disk_partitions()
      for usb_info_inst in usb_endpoint_infos:
        for partition in partitions:
          if partition.device == usb_info_inst.disk:
            return partition.mountpoint
    return None

  def add_log_note(self, msg):
    """Adds note in device logs if method available."""
    if self.add_log_note_func:
      self.add_log_note_func(msg)

  def copy_file_to(self, source_file_path, relative_destination_dir=None):
    """Copies source file to destination directory specified.

    Args:
        source_file_path (str): path to source file to copy to mass storage
          device
        relative_destination_dir (str): optional path relative to the mount
          path of the mass storage device

    Raises:
        DeviceError: if storage is not mounted, the mount path doesn't
        exist, or if the source file doesn't exist after copying the source file
        to the mount path + relative destination directory destination

    Note:
        The storage device must be mounted before calling this method. If the
        relative destination directory specified doesn't exist it will be
        created before the source file is copied
    """
    if not self.is_mounted():
      raise errors.DeviceError("Device copy file failed. "
                               "Device with Serial {} is not mounted.".format(
                                   self._serial_number))

    mount_point = self.mount_point
    if relative_destination_dir is not None:
      dst_file_path = os.path.join(mount_point, relative_destination_dir)
    else:
      dst_file_path = mount_point

    if not os.path.exists(dst_file_path):
      logger.debug("Creating destination path %s", dst_file_path)
      os.makedirs(dst_file_path)

    file_name = os.path.basename(source_file_path)
    dst_file_path = os.path.join(dst_file_path, file_name)
    msg = "Copying {} to {}".format(source_file_path, dst_file_path)
    logger.debug(msg)
    self.add_log_note(msg)
    shutil.copyfile(source_file_path, dst_file_path)
    if not os.path.exists(dst_file_path):
      raise errors.DeviceError(
          "Device copy file failed. "
          "File does not exist at destination {}"
          "Check destination path permissions.".format(dst_file_path))
    msg = "Copying {} to {} completed".format(source_file_path, dst_file_path)
    logger.debug(msg)
    self.add_log_note(msg)

  def eject(self, timeout=100.0):
    """Unmounts and ejects the mass storage device within the timeout specified.

    Args:
        timeout (float): time to wait to confirm the mass storage has been
          unmounted and ejected

    Raises:
        DeviceError: if timeout is exceeded before storage is confirmed as
        being unmounted or ejected.
    """
    start_time = time.time()
    self.add_log_note("{} eject requested.".format(self._serial_number))
    if self.is_device_ejected(timeout=.1):
      self.add_log_note("{} already ejected.".format(self._serial_number))
      return
    self.unmount(timeout=timeout)
    _validate_suid_bit(self.commands["EJECT"])
    cmd_list = self.commands["EJECT"].format(self.disk).split()
    try:
      subprocess.check_output(cmd_list)
    except subprocess.CalledProcessError as err:
      raise errors.DeviceError("Device eject failed. "
                               "{} produced this error: {!r}".format(
                                   cmd_list, err.output))
    if self.is_device_ejected(timeout=timeout):
      self.add_log_note("{} ejected in {}s".format(
          self._serial_number, int(time.time() - start_time)))
      return
    raise errors.DeviceError("Device eject failed. "
                             "Unable to eject USB mass storage {} in {}s"
                             "Check if device is currently in use. ".format(
                                 self.mount_point, timeout))

  def is_device_ejected(self, timeout=20):
    """Check whether device is ejected from the system.

    Args:
        timeout (int):  Number of seconds to wait for successful eject.

    Returns:
        bool: True if device is ejected.
    """
    exit_time = time.time() + timeout
    while time.time() < exit_time:
      devices_info = self._get_devices_info()
      if not devices_info:
        return True
      if sys.platform != "darwin":
        device_id_path = devices_info[0].disk
        with open("/etc/fstab", "r") as open_file:
          text = open_file.read()
        if not re.search(device_id_path, text):
          return True
      time.sleep(.5)
    return False

  def is_found(self):
    """Returns True if the mass storage device has been detected and is available to mount.

    Returns:
        bool: True if mass storage device is detected
    """
    return bool(self._get_devices_info())

  def is_mounted(self, timeout=20):
    """Returns True if the mass storage device is currently mounted.

    Args:
        timeout (float): time to wait for it to mount before timeout.

    Returns:
        bool: True if mass storage device is mounted

    Note:
        Uses a timeout as not all devices are using mount to mount the device
        and it can take a few seconds for the computers to register the external
        mount.
    """
    start_time = time.time()
    exit_time = start_time + timeout
    while time.time() < exit_time:
      if self.mount_point:
        msg = "{} found mounted at {} in {}s".format(
            self._serial_number, self.mount_point,
            int(time.time() - start_time))
        self.add_log_note(msg)
        logger.debug(msg)
        return True
    return False

  def mount(self, timeout=180):
    """Mounts the mass storage device within the timeout specified.

    Args:
        timeout (float): time to wait to confirm storage has been mounted
          and/or after manually attempting to mount the storage

    Raises:
        DeviceError: if timeout is exceeded before storage is confirmed as
        being mounted.

    Note:
        Since there is a udev rule that is responsible for performing the
        actual mounting of
        the device this method will wait up to the timeout specified before
        attempting to
        manually mount the device and then wait again for the timeout
        specified to confirm
        that the device has been successfully mounted
    """
    self.add_log_note("{} mount requested.".format(self._serial_number))
    if self.is_mounted():
      return
    if not self.is_found():
      self.add_log_note("Error: {} not found on machine.".format(
          self._serial_number))
      raise errors.DeviceError(
          "Device mount failed. "
          "USB mass storage matching Serial#={} was not found. "
          "\nCheck cables and device or try rebooting.".format(
              self._serial_number))
    msg = "Mounting {}".format(self.disk)
    logger.debug(msg)
    self.add_log_note(msg)
    _validate_suid_bit(self.commands["MOUNT"])
    cmd_list = self.commands["MOUNT"].format(self.disk).split()
    try:
      subprocess.check_output(cmd_list)
    except subprocess.CalledProcessError as err:
      if isinstance(err.output, bytes):
        err.output = err.output.decode("utf-8")
      if "Not authorized" in err.output:
        err = "Try running 'sudo chmod u+s {}' to fix permission error.".format(
            " ".join(cmd_list[0]))
      raise errors.DeviceError("Device mount failed. Err: {!r}".format(err))

    if not self.is_mounted(timeout=timeout):
      raise errors.DeviceError(
          "Device mount failed. Not mounted after {}s".format(timeout))

  def move_file_from(self,
                     relative_source_file_path,
                     destination,
                     delete_file=True):
    """Moves file from storage mount point to destination directory specified.

    Args:
        relative_source_file_path (str): relative path to file on device to
          copy over.
        destination (str): destination directory or file name path
        delete_file (bool): delete file from device.

    Raises:
        DeviceError: if storage is not mounted, the mount path doesn't
        exist, or if the source file doesn't exist

    Note:
        The storage device must be mounted before calling this method.
    """
    if not self.is_mounted():
      raise errors.DeviceError("Move_file_from failed. "
                               "Device with Serial {} is not mounted.".format(
                                   self._serial_number))

    source_file_path = os.path.join(self.mount_point, relative_source_file_path)

    if not os.path.exists(source_file_path):
      raise errors.DeviceError(
          "Device copy file failed. "
          "Source file {} doesn't exist. ".format(source_file_path))

    msg = "Copying {} to {}".format(source_file_path, destination)
    logger.debug(msg)
    self.add_log_note(msg)
    shutil.copy(source_file_path, destination)

    if delete_file:
      logger.debug("Removing %s", source_file_path)
      os.remove(source_file_path)

    if not os.path.exists(destination):
      raise errors.DeviceError(
          "Device copy file failed. "
          "Destination {} does not exist. "
          "Check destination path permissions.".format(destination))
    msg = "Copying {} to {} completed.".format(relative_source_file_path,
                                               destination)
    logger.debug(msg)
    self.add_log_note(msg)

  def unmount(self, timeout=180.0):
    """Unmounts the mass storage device within the timeout specified.

    Args:
        timeout (float): time to wait to confirm storage and successfully
          unmounted

    Raises:
        DeviceError: if timeout is exceeded before storage is confirmed as
        being unmounted
    """
    start_time = time.time()
    exit_time = start_time + timeout
    self.add_log_note("{} unmount requested".format(self._serial_number))
    mount_point = self.mount_point
    if not mount_point:
      self.add_log_note("{} already unmounted".format(self._serial_number))
      return
    self.add_log_note("{} unmounting from {}".format(self._serial_number,
                                                     mount_point))

    _validate_suid_bit(self.commands["UNMOUNT"])
    cmd_list = self.commands["UNMOUNT"].format(mount_point).split()
    try:
      subprocess.check_output(cmd_list)
    except subprocess.CalledProcessError as err:
      raise errors.DeviceError("Device unmount failed. "
                               "Error when calling /bin/umount: {}".format(
                                   err.output))

    while time.time() < exit_time:
      if not self.mount_point:
        msg = "{} successfully unmounted in {}s".format(
            self._serial_number, int(time.time() - start_time))
        logger.debug(msg)
        self.add_log_note(msg)
        return

    raise errors.DeviceError(
        "Device unmount failed. "
        "Unmounting USB mass storage at {} failed after {}s".format(
            self.mount_point, timeout))

  def _get_devices_info(self):
    """Returns info for the device.

    Returns:
      list: one or more matching usbinfo objects for each usb endpoint that
      matches.
    """
    return usb_utils.find_matching_connections(
        {"serial_number": {
            "include_regex": self._serial_number
        }})



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the switch_power_usb_with_charge capability."""
from typing import Any, Callable, Dict

from gazoo_device import decorators
from gazoo_device import gdm_logger
from gazoo_device.capabilities import switch_power_usb_default

logger = gdm_logger.get_logger()

CHARGE = "charge"
OFF = "off"
SYNC = "sync"


class SwitchPowerUsbWithCharge(switch_power_usb_default.SwitchPowerUsbDefault):
  """Definition of the switch_power_usb_with_charge capability."""
  _REQUIRED_COMMANDS = ["SET_MODE", "GET_MODE"]
  _REQUIRED_REGEXS = ["GET_MODE_REGEX"]

  def __init__(self,
               shell_fn: Callable[..., Any],
               regex_shell_fn: Callable[..., Any],
               command_dict: Dict[str, str],
               regex_dict: Dict[str, str],
               device_name: str,
               serial_number: str,
               total_ports: int):
    """Initializes an instance of SwitchPowerUsbWithCharge capability.

    Args:
      shell_fn: shell() function of the device.
      regex_shell_fn: shell_with_regex() function of the device.
      command_dict: A dictionary containing the command used for each method.
      regex_dict: Response regexes to use to when identifying command responses.
      device_name: name of the device this capability is attached to.
      serial_number: serial number of device this capability is attached to.
      total_ports: Number of ports on the device.
    """
    super().__init__(
        shell_fn=shell_fn,
        regex_shell_fn=regex_shell_fn,
        command_dict=command_dict,
        regex_dict=regex_dict,
        device_name=device_name,
        serial_number=serial_number,
        total_ports=total_ports)

  @decorators.PersistentProperty
  def supported_modes(self):
    """Get the USB power modes supported by the USB hub."""
    return [OFF, SYNC, CHARGE]

  def get_mode(self, port):
    """Gets the USB mode for the specified port.

    Args:
      port (int): Use this port to get the mode.

    Returns:
        str: USB port mode settings
             Example, 'off', 'sync', 'charge'

    Raises:
      DeviceError: invalid port.
    """
    port = int(port)
    self._validate_port("get_mode", port)
    flags = self._regex_shell_fn(
        self._command_dict["GET_MODE"].format(port),
        self._regex_dict["GET_MODE_REGEX"],
        tries=5)

    if "O" in flags:
      mode = OFF
    elif "S" in flags:
      mode = SYNC
    else:
      mode = CHARGE
    return mode

  @decorators.CapabilityLogDecorator(logger, decorators.DEBUG)
  def power_off(self, port):
    """This command powers off the port specified.

    Args:
        port (int): Identifies which hub port to power off.
    """
    port = int(port)
    self._validate_port("power_off", port)
    self.set_mode(OFF, port)

  @decorators.CapabilityLogDecorator(logger, decorators.DEBUG)
  def power_on(self, port, data_sync=True):
    """This command powers on the port specified.

    Args:
        port (int): Identifies which hub port to power on.
        data_sync (bool): True if data should be enabled, false for power
          only
    """
    port = int(port)
    self._validate_port("power_on", port)
    if data_sync:
      self.set_mode(SYNC, port)
    else:
      self.set_mode(CHARGE, port)

  @decorators.CapabilityLogDecorator(logger, decorators.DEBUG)
  def set_mode(self, mode, port):
    """Sets the given USB port to the mode specified.

    Args:
      mode (str): USB mode to set. Options: 'off', 'sync', 'charge'
      port (int): The port to set.

    Raises:
      DeviceError: invalid port, or mode.
    """
    port = int(port)
    self._validate_port("set_mode", port)
    self._validate_mode(mode)
    logger.debug("{} setting power mode to {} for usb port {}".format(
        self._device_name, mode, port))
    self._shell_fn(self._command_dict["SET_MODE"].format(mode, port))



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Cambrionix smart hub module.

This module contains APIs for interfacing with a Cambrionix smart hub device.

The following Cambrionix devices are currently supported::

    PP8S - A commercial quality 8 port smart hub device
    PP15S - A commercial quality 15 port smart hub device
    U16S - An uncased 16 port smart hub device

PP8S - Also known as the PowerPad 8S has 8 ports diagrammed below::

    XX      xxx
    Switch  Power
    xxx Host Port    --- Empty
    --- Empty        --- Empty
    --- Empty        --- Empty
    --- Empty        --- Empty
    xxx Port 1       xxx Port 5
    xxx Port 2       xxx Port 6
    xxx Port 3       xxx Port 7
    xxx Port 4       xxx Port 8

PP15S - Also known as the PowerPad 15S has 15 ports diagrammed below::

    XX      xxx
    Switch  Power

    xxx Host Port    xxx Port 8
    xxx Port 1       xxx Port 9
    xxx Port 2       xxx Port 10
    xxx Port 3       xxx Port 11
    xxx Port 4       xxx Port 12
    xxx Port 5       xxx Port 13
    xxx Port 6       xxx Port 14
    xxx Port 7       xxx Port 15

U16S - An uncased 16 port OEM device::

    xxx     xxx     xxx     xxx
    Port 1  Port 2  Port 3  Port 4                     xxx Host Port

    xxx     xxx     xxx     xxx                         X GND
    Port 5  Port 6  Port 7  Port 8                      X Power
                                                        X GPIO Pins?
    xxx     xxx     xxx     xxx                         X
    Port 9  Port 10 Port 11 Port 12                     X

    xxx     xxx     xxx     xxx         xxx          xxx
    Port 13 Port 14 Port 15 Port 16   Expansion 1  Expansion 2

Each of the above devices offers a serial port that can be used to control
the Cambrionix device. The code below accesses this serial port to allow
for controlling the device. Access the serial port using the following
serial port parameters::

    Baud Rate: 115200
    Data Bits: 8
       Parity: None
    Stop Bits: 1

Typical commands::

    help   - Display all available commands
    health - Display the health of the Cambrionix device
    l      - Live view of all active ports (also allows simple toggling of
    ports)
    mode   - Allows changing mode of a particular port (o-off, c-charge, s-sync)
    state  - Display the state of a particular port
    system - Display system information of the Cambrionix device

"""
import fcntl
import os
import select
import time
import typing

from gazoo_device import decorators
from gazoo_device import detect_criteria
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.base_classes import auxiliary_device
from gazoo_device.capabilities import switch_power_usb_with_charge
from gazoo_device.utility import deprecation_utils
from gazoo_device.utility import usb_config
from gazoo_device.utility import usb_utils
import serial

logger = gdm_logger.get_logger()

COMMANDS = {
    "CLEAR_ERROR_FLAGS": "cef",
    "CLEAR_REBOOTED_FLAG": "crf",
    "LIMITS": "limits",
    "STATE": "state",
    "GET_MODE": "state {}",
    "SYSTEM_STATUS": "system",
    "REBOOT_WATCHDOG": "reboot watchdog",
    "REBOOT": "reboot",
    "SET_MODE": "mode {} {}"
}

REGEXES = {
    "GET_MODE_REGEX": r"\s*\d{1,2},\s+\w+,\s+([A-Z0-9\s]+),",
}

TIMEOUTS = {
    "OPEN":
        30,  # Waits for other interactions. (reboot with watchdog=True takes ~16
    # seconds and is likely the longest running command)
    "PING": 3,
    "REBOOT": 3,
    "REBOOT_WATCHDOG": 15
}

_REBOOT_METHODS = ["watchdog", "shell"]


class Cambrionix(auxiliary_device.AuxiliaryDevice):
  """This class serves as a Python interface to a Cambrionix hub.

  Attributes: Hub Device Base Class
  """
  COMMUNICATION_TYPE = "SerialComms"
  DETECT_MATCH_CRITERIA = {
      detect_criteria.SerialQuery.product_name:
          "(ft230x basic uart)|(ps15-usb3)"
  }
  DEVICE_TYPE = "cambrionix"
  _OWNER_EMAIL = "gdm-authors@google.com"

  def __init__(self, manager, device_config, log_file_name, log_directory):
    """Constructor of the class Cambrionix.

    Args:
        manager(manager.Manager): Manager object for this device instance.
        device_config(dict): dict of two dicts, 'optional' device parameters
          and 'persistent' device parameters.
        log_file_name(str): file name in the log directory for device logs.
        log_directory(str): directory in which the controller will create
          the log file.

    Raises:
        DeviceError: unknown type of cambrionix
    """
    super().__init__(
        manager,
        device_config,
        log_file_name=log_file_name,
        log_directory=log_directory)
    self._commands.update(COMMANDS)
    self._regexes.update(REGEXES)
    self._timeouts.update(TIMEOUTS)
    self._serial_port = None

  @decorators.health_check
  def check_clear_flags(self):
    """clear the error and reboot flags."""
    self._command(self.commands["CLEAR_ERROR_FLAGS"])
    self._command(self.commands["CLEAR_REBOOTED_FLAG"])

  @decorators.DynamicProperty
  def firmware_version(self):
    """Gets the firmware version of the hub.

    Returns:
      str: The firmware version
    """
    return self._get_system_status()["firmware"]

  @decorators.PersistentProperty
  def health_checks(self):
    """Returns list of methods to execute as health checks."""
    return [self.check_device_connected, self.check_clear_flags]

  @decorators.PersistentProperty
  def port_extension_map(self):
    """Map of usb port extensions to port numbers."""
    return usb_config.CAMBRIONIX_PORT_MAP[self.model]

  @decorators.PersistentProperty
  def total_ports(self):
    """Number of usb ports for this device."""
    return len(self.port_extension_map.keys())

  @decorators.LogDecorator(logger, level=decorators.DEBUG)
  def _close(self):
    """Closes the serial port connection."""
    if self._serial_port is not None and self._serial_port.is_open:
      self._serial_port.close()

    super()._close()

  @decorators.PersistentProperty
  def communication_address(self):
    """Returns the path to the UART communication port."""
    return (self.props["persistent_identifiers"].get("console_port_name")
            or self.props["persistent_identifiers"]["hub_port_name"])

  @decorators.LogDecorator(logger)
  def get_detection_info(self):
    """Gets the persistent and optional attributes of a Cambrionix.

    Returns:
      tuple: (Dict of persistent attributes, dict of optional attributes)

    Raises:
      DeviceError: if device model is not supported.

    Notes:
      persistent: model,
                  console_port_name,
                  total_ports,
                  ftdi_serial_number,
                  serial_number

      optional:   empty dict
    """
    persistent_dict = self.props["persistent_identifiers"]
    persistent_dict["model"] = self._get_system_hardware()
    if persistent_dict["model"] not in usb_config.CAMBRIONIX_PORT_MAP:
      raise errors.DeviceError(
          "Model {} not supported. Supported models: {}".format(
              persistent_dict["model"],
              ",".join(usb_config.CAMBRIONIX_PORT_MAP.keys())))
    persistent_dict["console_port_name"] = self.communication_address
    persistent_dict["total_ports"] = self.total_ports
    persistent_dict[
        "ftdi_serial_number"] = usb_utils.get_serial_number_from_path(
            self.communication_address)

    # Cambrionix does not have a separate serial number from the one shown
    # in the /dev/serial/by-id/... name.
    persistent_dict["serial_number"] = self.props["persistent_identifiers"][
        "ftdi_serial_number"]

    self.props["options"] = {}

    return persistent_dict, self.props["options"]

  @classmethod
  def is_connected(cls, device_config):
    """Returns true if a Cambrionix is connected to the computer, False otherwise.

    Args:
      device_config (dict): contains "persistent" dict

    Returns:
      bool: Cambrionix is connected

    Notes:
      To return True, the cambrionix_config needs to have
      "console_port_name" in its
      "persistent" dict and that path needs to exist on the computer.
    """
    if "console_port_name" in device_config["persistent"]:
      address = device_config["persistent"]["console_port_name"]
    else:
      address = device_config["persistent"]["hub_port_name"]
    return os.path.exists(address)

  @decorators.LogDecorator(logger)
  def reboot(self, no_wait=False, method="shell"):
    """Reboots the system.

    Args:
      no_wait (bool): Return before boot up is complete. Default: False.
      method (str): Method to invoke reboot. 'shell' | 'watchdog'

    Raises:
      ValueError: if method name not recognized.

    Note:
      (Terminal Command Reference v1.5 page 15).
    """
    if method.lower() not in _REBOOT_METHODS:
      raise ValueError(
          f"Method {method!r} not recognized. Supported methods: {_REBOOT_METHODS}"
      )
    if method == "watchdog":
      self._command(
          self.commands["REBOOT_WATCHDOG"],
          close_delay=self.timeouts["REBOOT_WATCHDOG"])
    else:
      self._command(
          self.commands["REBOOT"], close_delay=self.timeouts["REBOOT"])

  @decorators.LogDecorator(logger)
  def recover(self, error):
    """Attempts to recover device based on the type of error specified.

    Note: The check_device_ready method can raise a number of separate
    exceptions which are passed to this method as exception objects. The
    recovery method is chosen based on the type of the error. A list of errors
    and recovery steps follow (not all recovery steps have been implemented
    yet):

    DeviceNotResponsiveError:
        * Recovery: reboot device.

    Args:
        error (CheckDeviceReadyError): A subclass of CheckDeviceReadyError
        that will be used to identify a possible recovery solution to use.

    Raises:
        DeviceError: If device recovery fails while attempting to perform
        recovery steps.
        CheckDeviceReadyError: If there are no recovery steps available for
        the error argument, it will be re-raised directly.
    """

    if isinstance(error, errors.DeviceNotResponsiveError):
      self.reboot()
    else:
      raise error

  def shell(self,
            command,
            command_name="shell",
            timeout=None,
            port=0,
            include_return_code=False):
    """Sends command and returns response.

    Args:
        command (str): Command to send to the device.
        command_name (str): Optional identifier to use in logs for this
          command.
        timeout (float): Seconds to wait for pattern after command sent.
        port (int): not used.
        include_return_code (bool): flag indicating return code should be
          returned

    Returns:
        str or tuple: response or (response, return_code) if
        include_return_code is True.
            return_code 0 - success
            return_code -1 - error
    """
    timeout = timeout or self.timeouts["SHELL"]
    response = ""
    if include_return_code:
      return_code = 0
      try:
        response = self._command(command)
        response = self._list_to_str(response)
      except errors.DeviceError:
        return_code = -1
      return response, return_code
    response = self._command(command)
    return self._list_to_str(response)

  def shell_with_regex(self,
                       command,
                       regex,
                       regex_group=1,
                       command_name="shell",
                       raise_error=False,
                       tries=1,
                       port=0):
    """Sends a command, searches for a regex in the response, and returns a match group.

    Args:
        command (str): command to issue.
        regex (str): regular expression with one or more capturing groups.
        regex_group (int): number of regex group to return.
        command_name (str): command name to appear in log messages.
        raise_error (bool): whether or not to raise error if unable to find a
          match.
        tries (int): how many times to try executing the command before failing.
        port (int): which port to send the shell command to.

    Returns:
        str: value of the capturing group with index 'regex_group' in the match.

    Raises:
        DeviceError: if command execution fails OR couldn't find the requested
        group in any of the responses.
    """
    return self.command_with_regex(
        command,
        regex,
        self.shell,
        regex_group=regex_group,
        raise_error=raise_error,
        tries=tries,
        command_name=command_name,
        port=port)

  @decorators.CapabilityDecorator(
      switch_power_usb_with_charge.SwitchPowerUsbWithCharge)
  def switch_power(self):
    """Returns a switch_power object to send commands.

    Returns:
        object: switch_power_usb_with_charge
    """
    return self.lazy_init(
        switch_power_usb_with_charge.SwitchPowerUsbWithCharge,
        shell_fn=self.shell,
        regex_shell_fn=self.shell_with_regex,
        command_dict=self.commands,
        regex_dict=self.regexes,
        device_name=self.name,
        serial_number=self.serial_number,
        total_ports=self.total_ports)

  @decorators.PersistentProperty
  def valid_modes(self):
    return ["off", "sync", "charge"]

  def _command(self, command, close_delay=0.0):
    """Opens the control serial port and sends a command.

    Args:
      command (str): Command to send to device
      close_delay (float): Seconds to delay between command and closing
        control port

    Returns:
      list: All strings in the response (except the trailing >> prompt)

    Raises:
      DeviceError: Error in response to command.

    Note:
      Returns response. Delays closing the port afterwards if close_delay >
      0.

      With some commands (e.g. reboot), it is necessary to wait before
      closing the control serial port to prevent other GDM instances from
      accessing the control serial port.
    """
    self._open()
    try:
      self.__write_command(self._serial_port, command)
      if command.startswith("reboot"):
        return

      response = self.__get_response(self._serial_port)
      if response[0].startswith("*E"):
        raise errors.DeviceError("Device {} command failed. "
                                 "Unable to write command: {} "
                                 "to serial port: {}  Err: {!r}".format(
                                     self.name, command, self._serial_port,
                                     response[0]))
    finally:
      if close_delay > 0.0:
        time.sleep(close_delay)
      self._serial_port.close()

    # Discard the last line which is the prompt
    return response[:-1]

  def _get_system_status(self):
    """Gets hardware and firmware information.

    Returns:
      dict: Information regarding the system

    Note:
      (Terminal Command Reference v1.5 page 6)
      Example output from PP8S device:
      >> system
      cambrionix PP8S 8 Port USB Charge+Sync
      Hardware: PP8S
      Firmware: 1.68
      Compiled: Feb 14 2017 17:30:26
      Group: -
      Panel ID: Absent
    """
    sysinfo_strings = self._command(self.commands["SYSTEM_STATUS"])
    sysinfo_dict = {"name": sysinfo_strings[0]}  # pytype: disable=unsupported-operands
    for line in sysinfo_strings:  # pytype: disable=attribute-error
      if ":" in line:
        key, value = line.split(":", 1)
        sysinfo_dict[key.lower()] = value.strip()

    return sysinfo_dict

  def _get_system_hardware(self):
    """Gets the hardware description of the hub.

    Returns:
      str: Hardware description
    """
    return self._get_system_status()["hardware"]

  def _list_to_str(self, lst):
    if isinstance(lst, list):
      lst = " ".join(map(str, lst))
    return lst

  def _open(self):
    start_time = time.time()
    error = ""
    while time.time() - start_time < self.timeouts["OPEN"]:
      try:
        if self._serial_port is None:
          # Windows is not supported due to use of fcntl.
          # Cast to Posix serial so pytype understands this.
          self._serial_port = typing.cast(
              serial.serialposix.Serial,
              serial.Serial(port=self.communication_address,
                            baudrate=115200,
                            timeout=0.1,
                            exclusive=True))
          # NEP-1852: Prevent inheritance of file descriptors to exec'd child
          # processes.
          file_descriptor = self._serial_port.fd
          flags = fcntl.fcntl(file_descriptor, fcntl.F_GETFD)
          fcntl.fcntl(file_descriptor, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)
          return
        if not self._serial_port.is_open:
          self._serial_port.open()
        return
      except Exception as err:
        error = err
    raise errors.DeviceError("Device {} open failed. "
                             "Unable to open control serial port in {} seconds"
                             "Error: {}".format(self.name,
                                                self.timeouts["OPEN"], error))

  @staticmethod
  def __write_command(serial_port, command):
    """Internal helper for writing to the hub.

    Args:
      serial_port (str): Cambrionix serial port.
      command (str): command to send to device.

    Raises:
      DeviceError: no response from device.
    """
    line_ending = "\r\n"
    ctrl_c_cmd = "\x03" + line_ending

    # Clear any existing text by sending a CTRL-C
    # command and waiting for a prompt
    serial_port.write(ctrl_c_cmd.encode("utf-8"))
    Cambrionix.__get_response(serial_port)

    if not command.endswith(line_ending):
      command += line_ending

    for char in command:
      serial_port.write(char.encode("utf-8"))
      if command.startswith("reboot") and char == "\r":
        break

      while True:
        ready = select.select([serial_port], [], [], 25)[0]
        if ready:
          if serial_port.read(1).decode("utf-8") == char:
            break
        else:
          raise errors.DeviceError("Device cambrionix write command failed. "
                                   "Read timeout on serial port: {} "
                                   "while writing command: {}".format(
                                       serial_port, command))

  @staticmethod
  def __get_response(serial_port):
    """Internal helper returning the response from the hub as a list of lines.

    Args:
      serial_port (str): Cambrionix serial port.

    Returns:
      str: response.

    Raises:
      DeviceError: Device not responding.

    Note:
      The command prompt is always this string: ">> ".
    """
    read_data = ""
    while not read_data.endswith("\n>> "):
      ready = select.select([serial_port], [], [], 25)[0]
      if ready:
        read_data += serial_port.read(serial_port.inWaiting()).decode(
            "utf-8", "replace")
      else:
        raise errors.DeviceError(
            "Device cambrionix get response failed. "
            "Read timeout on serial port: {}".format(serial_port))

    return read_data.splitlines()


deprecation_utils.add_deprecated_attributes(
    Cambrionix, [("set_mode", "switch_power.set_mode", True),
                 ("power_on", "switch_power.power_on", True),
                 ("power_off", "switch_power.power_off", True)])



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Fastboot capability interface."""
import abc
from gazoo_device.capabilities.interfaces import capability_base


class FastbootBase(capability_base.CapabilityBase):
  """Abstract base class for the Fastboot capability."""

  @abc.abstractmethod
  def enter(self):
    """Enters fastboot (bootloader) mode for updating device firmware.

    Raises:
        DeviceError: if unable to boot device into fastboot mode.
    """

  @abc.abstractmethod
  def exit(self):
    """Exits fastboot (bootloader) mode after updating device firmware.

    Raises:
        DeviceError if unable to reboot device out of fastboot mode
    """



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ExpectResponse object carries results from calls to GDM expect API.

GDM expect methods, found in switchboard.py, load results into ExpectResponse
to permit calling routines a way to evaluate device execution and behavior.
"""


class ExpectResponse(object):
  """Stores the values of expect Attributes.

  Attributes:
      index (int): index of matching pattern in pattern list (None if
        timeout).
      before (str): all the characters looked at before the match.
      after (str): all the characters after the first matching character.
      match (re.Match): pattern match object.
      timedout (bool): True if the expect timed out.
      time_elapsed (float): number of seconds between start and finish.
      remaining (list): patterns that have not been matched (after a timeout).
      match_list (list): of re.search Match objects.
  """

  def __init__(self,
               index,
               before,
               after,
               match,
               time_elapsed,
               timedout=False,
               remaining=None,
               match_list=None):
    self.index = index
    self.before = before
    self.after = after
    self.match = match
    self.timedout = timedout
    self.time_elapsed = time_elapsed
    self.remaining = [] if remaining is None else remaining
    self.match_list = [] if match_list is None else match_list



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: locking_service.proto
# pylint: skip-file
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='locking_service.proto',
  package='chip.rpc',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x15locking_service.proto\x12\x08\x63hip.rpc\"\x07\n\x05\x45mpty\"\x1e\n\x0cLockingState\x12\x0e\n\x06locked\x18\x01 \x01(\x08\x32m\n\x07Locking\x12\x30\n\x03Set\x12\x16.chip.rpc.LockingState\x1a\x0f.chip.rpc.Empty\"\x00\x12\x30\n\x03Get\x12\x0f.chip.rpc.Empty\x1a\x16.chip.rpc.LockingState\"\x00\x62\x06proto3'
)




_EMPTY = _descriptor.Descriptor(
  name='Empty',
  full_name='chip.rpc.Empty',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=35,
  serialized_end=42,
)


_LOCKINGSTATE = _descriptor.Descriptor(
  name='LockingState',
  full_name='chip.rpc.LockingState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='locked', full_name='chip.rpc.LockingState.locked', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=44,
  serialized_end=74,
)

DESCRIPTOR.message_types_by_name['Empty'] = _EMPTY
DESCRIPTOR.message_types_by_name['LockingState'] = _LOCKINGSTATE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Empty = _reflection.GeneratedProtocolMessageType('Empty', (_message.Message,), {
  'DESCRIPTOR' : _EMPTY,
  '__module__' : 'locking_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.Empty)
  })
_sym_db.RegisterMessage(Empty)

LockingState = _reflection.GeneratedProtocolMessageType('LockingState', (_message.Message,), {
  'DESCRIPTOR' : _LOCKINGSTATE,
  '__module__' : 'locking_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.LockingState)
  })
_sym_db.RegisterMessage(LockingState)



_LOCKING = _descriptor.ServiceDescriptor(
  name='Locking',
  full_name='chip.rpc.Locking',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=76,
  serialized_end=185,
  methods=[
  _descriptor.MethodDescriptor(
    name='Set',
    full_name='chip.rpc.Locking.Set',
    index=0,
    containing_service=None,
    input_type=_LOCKINGSTATE,
    output_type=_EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Get',
    full_name='chip.rpc.Locking.Get',
    index=1,
    containing_service=None,
    input_type=_EMPTY,
    output_type=_LOCKINGSTATE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_LOCKING)

DESCRIPTOR.services_by_name['Locking'] = _LOCKING

# @@protoc_insertion_point(module_scope)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Testbed module.

Usage:

To enable a testbed health check, you will need to...

1. Add a health check method to the Testbed class below. This method should
follow the naming convention 'check_<description>', be decorated with
'@decorators.health_check', and accept 1 argument which will be the testing
value of the property used to trigger the check (detailed in 2. below). You will
have access to the devices in your testbed via self.devices within the health
check method. If you'd like to issue a shell command to the device within the
method, add the shell command to the global COMMANDS dictionary below and use it
in the method via self.commands.

2. Add an entry to the 'self._prop_to_health_check' dictionary in
Testbed.__init__() below. The entry should follow the format
{
  <testing property>: <health check method>
}
The testing property should be the name of the property used to trigger the
health check, and the health check method should be the newly added Testbed
method. This will tell GDM to execute the method when the property exists in the
testbed config (detailed in 3. below).

3. Add the testing property from 2. above to the testbed configs of the testbeds
you'd like to execute the health check on.

Note that the property value will be passed to the corresponding health check
method as an argument. If no property value is necessary, use 'true'.
"""
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger

logger = gdm_logger.get_logger()

COMMANDS = {"FILE_SYSTEMS": "df"}


class Testbed(object):
  """Class for testbed interaction."""

  def __init__(self, devices, testing_props, name="testbed"):
    """Initialize a Testbed object.

    Args:
        devices (list): list of device instances.
        testing_props (dict): properties of the testbed used for testing.
        name (str): optional name of the testbed, defaulted to "testbed".
    """
    self.name = name
    self.devices = devices
    self.testing_props = testing_props
    self.commands = COMMANDS

    # map from testing prop to corresponding testbed health check method
    # e.g. if { "external_storage": self.check_external_storage_mounted }, then
    #      the check_external_storage_mounted method will be executed as a
    #      health check if the "external_storage" prop is provided during
    #      Testbed initialization.
    self._prop_to_health_check = {
        "external_storage": self.check_external_storage_mounted,
    }

  @decorators.health_check
  def check_external_storage_mounted(self, enabled):
    """Check if external USB storage is mounted on devices in the testbed.

    Args:
        enabled (str): whether to run this health check, "true" or "false".

    Raises:
        TestbedMissingExternalStorageError: if devices are missing external
        USB storage.
    """
    if enabled.lower() == "false":
      return

    file_system_cmd = self.commands["FILE_SYSTEMS"]
    usb_storage_location = "/data/usb"
    devices_missing_external_storage = []
    for device in self.devices:
      file_system_output = device.shell(file_system_cmd)
      file_system_mounts = self._get_filesystem_mounts(file_system_output)
      if usb_storage_location not in file_system_mounts:
        devices_missing_external_storage.append(device.name)

    if devices_missing_external_storage:
      raise errors.TestbedMissingExternalStorageError(
          device_names=devices_missing_external_storage,
          details=(
              f"command {file_system_cmd!r} did not display any filesystems "
              f"mounted on {usb_storage_location!r}"),
          recovery="mount external USB storage")

  @decorators.LogDecorator(logger)
  def check_testbed_ready(self):
    """Checks if the testbed is healthy by executing a series of health check methods."""
    health_check_call = []  # list of tuples: (health check method, prop value)
    for prop in self.testing_props:
      if prop in self._prop_to_health_check:
        health_check_call.append(
            (self._prop_to_health_check[prop], self.testing_props[prop]))
    self._execute_health_check_methods(health_check_call)

  @decorators.LogDecorator(logger)
  def make_testbed_ready(self, setting="on"):
    """Check testbed readiness and attempt recovery if allowed.

    Args:
        setting (str): 'on'|'off'|'check_only'.

    Raises:
        DeviceError: if testbed health checks fail.
    """
    if setting not in ["on", "off", "check_only"]:
      raise errors.DeviceError(
          "{} failed make_testbed_ready. Setting '{}' is invalid.".format(
              self.name, setting))
    elif setting == "off":
      return

    # check testbed health
    try:
      self.check_testbed_ready()
    except errors.DeviceError as err:
      if setting == "check_only":
        logger.info("{} skipping recovery".format(self.name))
        raise

      logger.info("{} failed check_testbed_ready with {}".format(
          self.name, repr(err)))

      # attempt to recover the testbed
      self.recover(err)
      logger.info("{} re-checking readiness after recovery attempt".format(
          self.name))
      self.check_testbed_ready()
      logger.info("{} successfully recovered to ready state".format(self.name))

  @decorators.LogDecorator(logger)
  def recover(self, error):
    """Attempts to recover testbed based on the type of error specified.

    Args:
        error (CheckTestbedReadyError): The exception thrown that will be
          used to identify a possible recovery solution to use.

    Raises:
        CheckTestbedReadyError: If recovery fails.
    """
    raise error

  def _execute_health_check_methods(self, health_check_calls):
    """Execute health checks on the testbed.

    Args:
        health_check_calls (list): list of tuples as (health check method,
          argument) to execute as health checks.

    Raises:
        CheckTestbedReadyError: if health check fails.

    Notes:
        - The value of the property that triggered the health check is
        passed to the health
        check method as an argument; therefore, health check methods will
        accept a single
        argument (unlike device health checks). All property values will be
        strings.
        - Some health checks will have "true" / "false" argument values. The
        health check
        should be skipped when the argument is "false".
        - Order of health check methods matter. Health check execution will
        stop at the first
        failing health check.
    """
    checks_passed = []
    for step, health_check_call in enumerate(health_check_calls):
      health_check_method, arg = health_check_call
      method_name = health_check_method.__name__
      health_check_name = method_name.replace("_", " ").strip().capitalize()

      try:
        health_check_method(arg)
      except errors.CheckTestbedReadyError as err:
        logger.info("testbed health check {}/{} failed: {}.".format(
            step + 1, len(health_check_calls), health_check_name))
        err.checks_passed = checks_passed
        err.properties = self.testing_props.copy()
        raise

      checks_passed.append("{}.{}".format(type(self).__name__, method_name))
      logger.info("testbed health check {}/{} succeeded: {}.".format(
          step + 1, len(health_check_calls), health_check_name))

  def _get_filesystem_mounts(self, file_system_output):
    """Parse locations that file systems are mounted on from command output.

    Args:
        file_system_output (str): output of command listing device file
          systems.

    Returns:
        list: locations that file systems are mounted on on the device.
    """
    file_system_mounts = []
    for file_system in file_system_output.splitlines():
      if not file_system:
        continue
      file_system_info = file_system.split()

      # 6 columns of information
      # Filesystem,1K-blocks,Used,Available,Use%,Mounted on
      if len(file_system_info) == 6:
        file_system_mounts.append(file_system_info[5])

    return file_system_mounts



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Gazoo Device Manager CLI.

The CLI is generated dynamically by Python Fire:
https://github.com/google/python-fire.
"""
import sys
from typing import Dict, Optional, Sequence

import fire

import gazoo_device
from gazoo_device import errors
from gazoo_device import extensions
from gazoo_device import fire_manager
from gazoo_device import fire_patch
from gazoo_device import gdm_logger
from gazoo_device import package_registrar

logger = gdm_logger.get_logger()

VERSION_FLAG = "-v"
FLAG_MARKER = "--"
OMIT_FLAGS = ["help"]
_CLI_NAME = "gdm"


def execute_command(command: Optional[str] = None,
                    cli_name: str = _CLI_NAME) -> int:
  """Executes the CLI command through Python Fire.

  Args:
    command: Passed to Python Fire. If None, sys.argv are used instead.
    cli_name: Name of the CLI executable ("gdm").

  Returns:
    Error code: 0 if command was successful, non-zero otherwise.
  """
  # Parse flags out of commands. E.g. "gdm --debug - devices" ->
  # flags = {"debug": True}, commands = ["-", "devices"].
  if command:
    args = command.split()
  else:
    args = sys.argv[1:]
  flags = _get_flags(args)
  commands = [arg for arg in args if arg[len(FLAG_MARKER):] not in flags.keys()]

  # Instantiate FireManager instance with provided flags
  manager_inst = fire_manager.FireManager(**flags)

  # Execute CLI command
  exit_code = 0
  try:
    fire_patch.apply_patch()
    fire.Fire(manager_inst, commands, name=cli_name)
  except (ValueError, errors.DeviceError) as err:
    logger.error((repr(err)))
    exit_code = 1
  except KeyboardInterrupt:
    exit_code = 2
  finally:
    manager_inst.close()

  return exit_code


def _get_flags(args: Sequence[str]) -> Dict[str, bool]:
  """Parses flags out of array of CLI args.

  Flags in OMIT_FLAGS dict will not be returned.

  Args:
    args: CLI arguments provided by the user.

  Returns:
    Parsed flags to pass to the CLI.
  """
  flags = {}
  for arg in args:
    if arg.startswith(FLAG_MARKER):
      flag_name = arg[len(FLAG_MARKER):]
      if flag_name and flag_name not in OMIT_FLAGS:
        flags[flag_name] = True
    else:
      break  # Ignore flags after initial CLI call
  return flags


def main(command: Optional[str] = None) -> int:
  """Main function for Gazoo Device Manager (gazoo_device) package.

  Args:
    command: Passed to Python Fire. If None, sys.argv are used instead.

  Returns:
    Error code: 0 if command was successful, non-zero otherwise.
  """
  package_registrar.import_and_register_cli_extension_packages()

  if VERSION_FLAG in sys.argv or (command and VERSION_FLAG in command):
    logger.info(f"Gazoo Device Manager {gazoo_device.version}")
    package_versions = extensions.get_registered_package_info()
    logger.info(f"Registered extension packages: {package_versions}")
    return 0

  return execute_command(command)


if __name__ == "__main__":
  main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Device class for ESP32 locking device."""
from gazoo_device import decorators
from gazoo_device import detect_criteria
from gazoo_device import gdm_logger
from gazoo_device.base_classes import espressif_esp32_device
from gazoo_device.capabilities import pwrpc_lock_default
from gazoo_device.protos import locking_service_pb2
from gazoo_device.utility import pwrpc_utils
import immutabledict


logger = gdm_logger.get_logger()
_LOCK_SERVICE_REGEXES = immutabledict.immutabledict({
    pwrpc_lock_default.LockedState.LOCKED: "Lock Action has been completed",
    pwrpc_lock_default.LockedState.UNLOCKED:
        "Unlock Action has been completed"})
_RPC_TIMEOUT = 10  # seconds


class ESP32PigweedLocking(espressif_esp32_device.EspressifESP32Device):
  """Device class for ESP32PigweedLocking devices.

  This device class is for the Pigweed locking application running on
  the Espressif ESP32 M5Stack platform:
  https://github.com/project-chip/connectedhomeip/tree/master/examples/lock-app/esp32
  """
  DETECT_MATCH_CRITERIA = {
      detect_criteria.PigweedQuery.product_name:
          "cp2104 usb to uart bridge controller",
      detect_criteria.PigweedQuery.manufacturer_name: "silicon_labs",
      detect_criteria.PigweedQuery.app_type:
          pwrpc_utils.PigweedAppType.LOCKING.value,
  }
  DEVICE_TYPE = "esp32pigweedlocking"
  _OWNER_EMAIL = "gdm-authors@google.com"
  _COMMUNICATION_KWARGS = {"protobufs": (locking_service_pb2,),
                           "baudrate": espressif_esp32_device.BAUDRATE}

  @decorators.DynamicProperty
  def firmware_version(self):
    """Firmware version of the device."""
    # TODO(b/193478488) Add pwrpc_common capability.
    return "NOT_IMPLEMENTED"

  @decorators.CapabilityDecorator(pwrpc_lock_default.PwRPCLockDefault)
  def pw_rpc_lock(self):
    """PwRPCLock instance to send RPC commands."""
    return self.lazy_init(
        pwrpc_lock_default.PwRPCLockDefault,
        device_name=self.name,
        expect_locking_regexes=_LOCK_SERVICE_REGEXES,
        expect_timeout=_RPC_TIMEOUT,
        switchboard_call=self.switchboard.call,
        switchboard_call_expect=self.switchboard.call_and_expect)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains APIs for interfacing with a Yepkit hub device.

Use the following commands to control the On/Off of the devices connected
to the YKUSH board downstream ports.

Power down a device connected to YKUSH board downstream port 1:

$ ykushcmd -d 1

Power up a device connected to YKUSH board downstream port 1:

$ ykushcmd -u 1

To find out the serial number of a YKUSH board, connect it to the host and
execute the command:

$ ykushcmd -l

Power down a device connected to YKUSH board downstream port 2 of YKB4388:

$ ykushcmd -s YKB4388 -d 2

Power up a device connected to YKUSH board downstream port 2 of YKB4388:

$ ykushcmd -s YKB4388 -u 2

To get the current switching status (UP/DOWN) of a downstream port 2.

$ ykushcmd -g 2
"""
import os
import subprocess

from gazoo_device import decorators
from gazoo_device import detect_criteria
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.base_classes import auxiliary_device
from gazoo_device.capabilities import switch_power_usb_default
from gazoo_device.utility import deprecation_utils
from gazoo_device.utility import host_utils

logger = gdm_logger.get_logger()

yepkit_enabled = not bool(os.system("which ykushcmd > /dev/null"))

MODEL = "yepkit"
PRODUCT_ID = "f2f7"
VENDOR_ID = "04d8"
YKUSH_PORT_MAP = {"1": 1, "2": 2, "3": 3}

COMMANDS = {
    "POWER_OFF": "ykushcmd -s {} -d {}",
    "POWER_ON": "ykushcmd -s {} -u {}",
    "GET_MODE": "ykushcmd -s {} -g {}",
    "LIST_YEPKITS": "ykushcmd -l"
}

REGEXES = {
    "GET_MODE_REGEX": r"is[ ]([A-Z]+)",
}

TIMEOUTS = {}


class Yepkit(auxiliary_device.AuxiliaryDevice):
  """This class serves as a Python interface to a Yepkit hub."""
  COMMUNICATION_TYPE = "YepkitComms"
  DETECT_MATCH_CRITERIA = {detect_criteria.GenericQuery.always_true: True}
  DEVICE_TYPE = "yepkit"
  _MODEL = MODEL
  _OWNER_EMAIL = "gdm-authors@google.com"

  def __init__(self,
               manager,
               device_config,
               log_file_name=None,
               log_directory=None):
    """Initialize the Yepkit class with the serial port information provided.

    Args:
        manager(manager.Manager): Manager object for this device instance.
        device_config (dict): dict of two dicts, 'optional' device parameters
          and 'persistent' device parameters.
        log_file_name (str): file name in the log directory for device logs.
        log_directory (str): directory in which the controller will create
          the log file.

    Raises:
        RuntimeError: if yepkit is not enabled.
    """
    if not yepkit_enabled:
      raise RuntimeError("yepkit is not enabled.")

    super().__init__(
        manager,
        device_config,
        log_file_name=log_file_name,
        log_directory=log_directory)
    self._commands.update(COMMANDS)
    self._regexes.update(REGEXES)
    self._timeouts.update(TIMEOUTS)

  @decorators.PersistentProperty
  def health_checks(self):
    """Returns list of methods to execute as health checks."""
    return [self.check_device_connected]

  @decorators.PersistentProperty
  def total_ports(self):
    """Number of ports on Yepkit."""
    return 3

  @decorators.CapabilityDecorator(switch_power_usb_default.SwitchPowerUsbDefault
                                 )
  def switch_power(self):
    """Returns a switch_power object to send commands.

    Returns:
        SwitchPowerUsbDefault: switch_power_usb_default
    """
    return self.lazy_init(
        switch_power_usb_default.SwitchPowerUsbDefault,
        shell_fn=self.shell,
        regex_shell_fn=self.shell_with_regex,
        command_dict=self.commands,
        regex_dict=self.regexes,
        device_name=self.name,
        serial_number=self.serial_number,
        total_ports=self.total_ports)

  @decorators.LogDecorator(logger)
  def get_detection_info(self):
    """Gets the persistent and optional attributes of a yepkit.

    Returns:
        tuple: (dict, dict) dictionary of persistent attributes,
               dictionary of optional attributes (set to None).

    Raises:
        RuntimeError: if yepkit is not enabled.

    Notes:
        persistent: device_type, model, console_port_name, serial_number,
                    total_ports

        optional: {}
    """
    if not yepkit_enabled:
      raise RuntimeError("yepkit is not enabled.")

    if len(self.communication_address) < 4:
      raise RuntimeError("{} serial number {} is too short. "
                         "Length of serial number must be >= 4".format(
                             self.DEVICE_TYPE, self.serial_number))

    self.props["options"] = {}

    persistent_dict = self.props["persistent_identifiers"]
    persistent_dict["serial_number"] = self.communication_address
    persistent_dict["model"] = self._MODEL
    persistent_dict["total_ports"] = 3

    return self.props["persistent_identifiers"], self.props["options"]

  @classmethod
  def is_connected(cls, device_config):
    """Checks whether or not the yepkit is connected to the computer.

    Args:
        device_config (dict): contains "persistent" dict

    Returns:
        bool: whether device serial_number is in
        host_utils.get_all_yepkit_serials

    Raises:
        RuntimeError: if yepkit is not enabled.
    """
    return device_config["persistent"][
        "console_port_name"] in host_utils.get_all_yepkit_serials()

  def shell(self,
            command,
            command_name="shell",
            timeout=None,
            port=0,
            include_return_code=False):
    """Sends command and returns response.

    Args:
        command (str): Command to send to the device.
        command_name (str): Optional identifier to use in logs for this command.
        timeout (float): Seconds to wait for pattern after command sent.
        port (int): no op.
        include_return_code (bool): flag indicating return code should be
          returned

    Returns:
        str or tuple: response or (response, return_code) if
        include_return_code is True.
            return_code 0 - success
            return_code -1 - error

    Raises:
        DeviceError: if command execution fails OR couldn't find the requested
                     group in any of the responses.
    """
    timeout = timeout or self.timeouts["SHELL"]
    if "ykushcmd" not in command:
      raise errors.DeviceError(
          "{} shell expected 'ykushcmd' instead found: {!r}".format(
              self.name, command))
    raw_list = command.split(" ")
    if include_return_code:
      result = subprocess.check_output(raw_list).decode("utf-8", "replace")
      if "Application Help" in result:
        return result, -1  # Error
      return result, 0
    return subprocess.check_output(raw_list).decode("utf-8", "replace")

  def shell_with_regex(self,
                       command,
                       regex,
                       regex_group=1,
                       command_name="shell",
                       raise_error=False,
                       tries=1,
                       port=0):
    """Sends a command, searches for a regex in the response, and returns a match group.

    Args:
        command (str): command to issue.
        regex (str): regular expression with one or more capturing groups.
        regex_group (int): number of regex group to return.
        command_name (str): command name to appear in log messages.
        raise_error (bool): whether or not to raise error if unable to find
          a match.
        tries (int): how many times to try executing the command before failing.
        port (int): which port to send the shell command to.

    Returns:
        str: value of the capturing group with index 'regex_group' in the match.

    Raises:
        DeviceError: if command execution fails OR couldn't find the requested
                     group in any of the responses.
    """
    return self.command_with_regex(
        command,
        regex,
        self.shell,
        regex_group=regex_group,
        raise_error=raise_error,
        tries=tries,
        command_name=command_name,
        port=port)


deprecation_utils.add_deprecated_attributes(
    Yepkit, [("power_on", "switch_power.power_on", True),
             ("power_off", "switch_power.power_off", True)])



"""Allows for triggering GPIO lines on FTDI based boards."""
import time

import pylibftdi


class FtdiButtons:
  """Allows for triggering GPIO lines on FTDI based boards."""

  def __init__(self, ftdi_serial_number, button_map, polarity=False):
    """Initializes the FtdiButtons with the properties provided.

    The ftdi_serial_number is needed to unique identify the FTDI device
    to control as there may be more than one FTDI device available. The
    button map defines the interfaces and pin (bit) to set for each button
    and should look like this:
    some_button_map = {"a_button": {"interface": 1, "pin": 3},
                       "b_button": {"interface": 2, "pin": 5}}

    Args:
        ftdi_serial_number (str): FTDI serial number to use
        button_map (dict): button to pin (bit) mapping.
        polarity (bool): GPIO pin polarity to use (True to push button).
    """
    self._button_map = button_map
    self._button_down = {}
    self._ftdi_serial_number = ftdi_serial_number
    self._polarity = polarity
    self._sub_process = None
    self._bb_map = {}
    self._bitbang_device_init = False
    self.bitbang_device_init()

  def __del__(self):
    self.close()

  def close(self):
    """Release any held buttons on close."""
    for button in list(self._button_down):
      self.release(button)

    for interface in self._bb_map:
      ftdi_device = self._bb_map[interface]
      ftdi_device.close()

    self._bb_map.clear()
    self._bitbang_device_init = False

  def is_valid(self, button):
    """Return True if button specified is in the list of valid buttons.

    Args:
        button (str): button identifier.

    Returns:
        bool: True if button specified is in button map provided
    """

    return button in list(self._button_map.keys())

  def bitbang_device_init(self):
    """Creates BitBangDevices for each button and initializes to 'off'."""
    for button in list(self._button_map.keys()):
      interface = self._button_map[button]["interface"]
      pin = self._button_map[button]["pin"]

      if interface not in self._bb_map:
        ftdi_device = pylibftdi.BitBangDevice(
            device_id=self._ftdi_serial_number,
            interface_select=interface,
            direction=(1 << pin))
        ftdi_device.port = 0 if self._polarity else 255
        self._bb_map[interface] = ftdi_device
      else:
        self._bb_map[interface].direction |= (1 << pin)
      self._button_down[button] = False

    self._bitbang_device_init = True

  def press(self, button, wait=0.0):
    """Presses button and waits for the time specified.

    Note: You must call release or close to release the button later.

    Args:
        button (str): identify which button to press
        wait (float): seconds to wait before returning

    Raises:
        ValueError: invalid button or wait given.
    """
    if not self._bitbang_device_init:
      self.bitbang_device_init()
    if button not in self._button_map:
      raise ValueError("Invalid button {} specified".format(button))
    elif wait < 0.0:
      raise ValueError("Invalid wait {} specified".format(wait))
    elif not self._button_down[button]:
      self._gpio(button, -1, active_high=self._polarity)
      time.sleep(wait)

  def click(self, button, duration=.5):
    """Presses the button specified and holds it for the specified duration.

    Args:
        button (str): button to click.
        duration (float): seconds to wait before releasing the button.

    Raises:
        ValueError: invalid button or duration given.
    """
    if not self._bitbang_device_init:
      self.bitbang_device_init()
    if button not in self._button_map:
      raise ValueError("Invalid button {} specified".format(button))
    elif duration <= 0.0:
      raise ValueError("Invalid duration {} specified".format(duration))
    else:
      self._gpio(button, duration, active_high=self._polarity)

  def release(self, button):
    """Releases the button specified if it was previously pressed.

    Args:
        button (str): button to click.

    Raises:
        ValueError: invalid button given.
    """
    if not self._bitbang_device_init:
      self.bitbang_device_init()
    if button not in self._button_map:
      raise ValueError("Invalid button {} specified".format(button))
    elif self._button_down[button]:
      self._gpio(button, -1, active_high=self._polarity)

  def read_pin(self, interface, pin):
    """Returns True if the bit is high for the given pin, False otherwise.

    Args:
        interface (int): the port on the FTDI chip containing the GPIO
        pin (int): the bit representing the GPIO to read
    Return:
       bool: True if pin is high, False if low.
    """
    if not self._bitbang_device_init:
      self.bitbang_device_init()
    return (self._bb_map[interface].port & (1 << pin)) != 0

  def valid_buttons(self):
    """Returns a list of valid button names.

    Returns:
        list: A list of valid button name strings
    """
    return list(self._button_map.keys())

  def _gpio(self, button, duration, active_high):
    """Drive the GPIO of the button specified for the duration provided.

    Args:
        button (str): button to be toggled
        duration (float): seconds to wait between the two toggles. If
          negative, only one toggle will be performed (either a press or
          release).
        active_high (bool): drive the pin high if True otherwise drive
          active_low.
    """
    if not self._bitbang_device_init:
      self.bitbang_device_init()
    interface = self._button_map[button]["interface"]
    bb = self._bb_map[interface]
    self.__toggle(button, bb, active_high)
    if duration >= 0.0:
      time.sleep(duration)
      self.__toggle(button, bb, active_high)

  def __toggle(self, button, bb, active_high):
    """Toggles the button specified based on its previous state.

    Args:
        button (str): button to be toggled
        bb (BitBangDevice): performs the toggle.
        active_high (bool): drive the pin high if True, otherwise drive
          active_low.
    """
    if not self._bitbang_device_init:
      self.bitbang_device_init()
    pin = self._button_map[button]["pin"]
    if self._button_down[button]:
      # Release button
      if active_high:
        self.__clear_pin(bb, pin)
      else:
        self.__set_pin(bb, pin)
      self._button_down[button] = False
    else:
      # Set button
      if active_high:
        self.__set_pin(bb, pin)
      else:
        self.__clear_pin(bb, pin)
      self._button_down[button] = True

  @staticmethod
  def __clear_pin(bb, pin):
    """Clear the bit of the given pin.

    Args:
        bb (BitBangDevice): performs the work
        pin (int): the bit representing the GPIO to clear
    """
    bb.port &= 255 ^ (1 << pin)

  @staticmethod
  def __set_pin(bb, pin):
    """Sets the bit for the given pin.

    Args:
        bb (BitBangDevice): performs the work
        pin (int): the bit representing the GPIO to clear
    """
    bb.port |= (1 << pin)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility module for HTTP GET and POST methods."""
import json
import os
import socket
import urllib

from gazoo_device import gdm_logger
import requests
from requests import adapters
from urllib3 import poolmanager
import urllib3.exceptions

MAX_URL_RETRY = 3

logger = gdm_logger.get_logger()


class SSLAdapter(adapters.HTTPAdapter):
  """An HTTPS Transport Adapter that uses an arbitrary SSL version."""

  def __init__(self, ssl_version=None, **kwargs):
    self.ssl_version = ssl_version

    super(SSLAdapter, self).__init__(**kwargs)

  def init_poolmanager(self, connections, maxsize, block=False):
    self.poolmanager = poolmanager.PoolManager(
        num_pools=connections,
        maxsize=maxsize,
        block=block,
        ssl_version=self.ssl_version)


def download_file(url, local_path):
  """Downloads the file to the local path.

  Args:
    url (str): URL to file.
    local_path (str): location to download file to.

  Raises:
    RuntimeError: if unable to access domain or to download file in three tries.
  """
  for _ in range(MAX_URL_RETRY):
    try:
      validate_url_access(url)
      (_, header) = urllib.request.urlretrieve(url, local_path)
    except Exception as err:
      raise RuntimeError("Unable to download {} "
                         "Error: {!r}".format(url, err))
    if "last-modified" in header and os.path.exists(local_path):
      logger.info("Successfully downloaded file to {}".format(local_path))
      return
  raise RuntimeError("Unable to download {} after {} tries.".format(
      url, MAX_URL_RETRY))


def read_raw_html_page(url):
  """Returns the raw text on the HTML page.

  Args:
     url (str): URL for page.

  Returns:
     str: raw text on the HTML page.

  Raises:
    RuntimeError: if unable to access domain or to download file in three tries.
  """
  code = 0
  for _ in range(MAX_URL_RETRY):
    try:
      validate_url_access(url)
      response = urllib.request.urlopen(url)
      code = response.getcode()
      raw_page = response.read().decode("utf-8")
      if code == 200:
        # get all relative links to files
        return raw_page
      else:
        err = raw_page
    except urllib.error.HTTPError as err:
      code = err.code  # retry
      raw_page = err
  raise RuntimeError(
      "Unable to get files at URL {}. HTTP return code {}: {}".format(
          url, code, raw_page))


def send_http_get(url,
                  auth=None,
                  params=None,
                  data=None,
                  headers=None,
                  ssl_version=None,
                  valid_return_codes=None,
                  timeout=10,
                  tries=1):
  """Issues a HTTP GET and returns the response if the request is successful.

  Args:
      url (str): HTTP URL to which HTTP GET request needs to be sent
      auth (auth.AuthBase): HTTP authentication object (i.e. HTTPDigestAuth)
      params (dict): Parameters to send as a query string.
      data (dict): Data that is needed for this HTTP GET Request
      headers (dict): Headers that is needed for this HTTP GET Request
      ssl_version (int): SSL version to be used for secure http (https)
        requests. For example, ssl.PROTOCOL_TLSv1_2.
      valid_return_codes (list): List of valid HTTP return codes.
      timeout (int): request timeout in seconds
      tries (int): how many times to try sending the request.

  Raises:
      RuntimeError: if response.status_code returned by requests.get() is not
      in
                    valid_return_codes
      TypeError: if headers is not a dictionary
      RequestException, HTTPError: raised if request fails and are converted
      into a RuntimeError.

  Returns:
      requests.Response: Response returned by requests.get
  """
  valid_return_codes = valid_return_codes or [200]

  if data and not isinstance(data, dict):
    raise TypeError("Expecting a dict value data param but received: {}".format(
        type(data)))
  if headers and not isinstance(headers, dict):
    raise TypeError(
        "Expecting a dict value in headers param but received: {}".format(
            type(headers)))
  try:
    session = requests.Session()
    if ssl_version:
      session.mount("https://", SSLAdapter(ssl_version))
    if data:
      data = json.dumps(data)

    for attempt in range(tries):
      try:
        response = session.get(
            url,
            auth=auth,
            params=params,
            data=data,
            headers=headers,
            timeout=timeout,
            verify=False)
        break
      except (requests.exceptions.RequestException,
              urllib3.exceptions.HTTPError) as err:
        if attempt < tries - 1:
          logger.debug(
              "HTTP GET to URL {} with data {} failed. Retrying. Error: {!r}"
              .format(url, data, err))
        else:
          raise

  except Exception as err:
    raise RuntimeError(
        "HTTP GET to URL {} with data {} failed with error: {!r}".format(
            url, data, err))

  if response.status_code not in valid_return_codes:
    raise RuntimeError("HTTP GET to URL {} returned: {}".format(
        url, response.reason))

  return response


def send_http_post(url,
                   auth=None,
                   data=None,
                   headers=None,
                   json_data=None,
                   ssl_version=None,
                   valid_return_codes=None,
                   timeout=10,
                   tries=1):
  """Issues a HTTP POST and returns the response if the request is successful.

  Args:
      url (str): HTTP URL to which HTTP POST request needs to be sent
      auth (auth.AuthBase): HTTP authentication object (i.e. HTTPDigestAuth)
      data (dict): Data that is needed for this HTTP GET Request
      headers (dict): Headers needed that is needed for this HTTP POST Request
      json_data (dict): JSON data that is needed for this HTTP POST Request
      ssl_version (int): SSL version to be used for secure http (https)
        requests. For example, ssl.PROTOCOL_TLSv1_2
      valid_return_codes (list): List of valid HTTP return codes.
      timeout (int): request timeout in seconds
      tries (int): how many times to try sending the request.

  Raises:
      RuntimeError: if response.status_code returned by requests.post() is not
      in
      valid_return_codes
      TypeError: if headers / json_data is not a dictionary or None.
      RequestException, HTTPError: raised if request fails and are converted
      into a RuntimeError.

  Returns:
      requests.Response: Response returned by requests.post
  """
  headers = headers or {}
  json_data = json_data or {}
  valid_return_codes = valid_return_codes or [200]

  if not isinstance(headers, dict) or not isinstance(json_data, dict):
    raise TypeError(
        "Expecting a dict value in headers and json_data params but received: "
        "{} {} respectively".format(type(headers), type(json_data)))

  try:
    session = requests.Session()
    if ssl_version:
      session.mount("https://", SSLAdapter(ssl_version))

    for attempt in range(tries):
      try:
        response = session.post(
            url,
            auth=auth,
            data=data,
            json=json_data,
            headers=headers,
            timeout=timeout,
            verify=False)
        break
      except (requests.exceptions.RequestException,
              urllib3.exceptions.HTTPError) as err:
        if attempt < tries - 1:
          logger.debug(
              "HTTP POST to URL {} with headers {}, data {} and json data {} "
              "failed. Retrying. Error: {!r}".format(url, headers, data,
                                                     json_data, err))
        else:
          raise

  except Exception as err:
    raise RuntimeError(
        "HTTP POST to URL {} with headers {}, data {} and json data {} failed "
        "with error: {!r}".format(url, headers, data, json_data, err))

  if response.status_code not in valid_return_codes:
    raise RuntimeError(
        "HTTP POST to URL {} with headers {}, data {} and json data {} "
        "returned: {}".format(url, headers, data, json_data, response.reason))

  return response


def is_valid_ip_address(address, is_ipv6=False):
  """Checks if valid IP address.

  Args:
      address (str): IP address
      is_ipv6 (bool): if IP address is IPv6

  Returns:
      bool: True if valid IP else False
  """
  address_family = socket.AF_INET6 if is_ipv6 else socket.AF_INET
  if is_ipv6 and "%" in address:  # Check for link local Ipv6 address
    # Example: fe80::123:1234:1234:123%en0
    address = address.split("%")[0]
  try:
    socket.inet_pton(address_family, address)
  except socket.error:  # not a valid address
    return False
  return True


def validate_url_access(url):
  """Validates URL access.

  Args:
    url (str): URL to access.

  Raises:
    RuntimeError: if unable to access domain of URL.
  """
  parsed_uri = urllib.parse.urlparse(url)
  prefix = "{uri.scheme}://{uri.netloc}/".format(uri=parsed_uri)
  response_code = urllib.request.urlopen(prefix).getcode()
  if response_code != 200:
    raise RuntimeError("Unable to reach domain {}. Returned code {}".format(
        prefix, response_code))



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Device class for nRF Pigweed Lighting device."""
from gazoo_device import decorators
from gazoo_device import detect_criteria
from gazoo_device import gdm_logger
from gazoo_device.base_classes import nrf_connect_sdk_device
from gazoo_device.capabilities import pwrpc_button_default
from gazoo_device.capabilities import pwrpc_common_default
from gazoo_device.capabilities import pwrpc_light_default
from gazoo_device.utility import pwrpc_utils
import immutabledict

# TODO(b/185956488): Remove conditional imports of Pigweed
try:
  # pylint: disable=g-import-not-at-top
  # pytype: disable=import-error
  from button_service import button_service_pb2
  from device_service import device_service_pb2
  from lighting_service import lighting_service_pb2
  # pytype: enable=import-error
except ImportError:
  button_service_pb2 = None
  device_service_pb2 = None
  lighting_service_pb2 = None

logger = gdm_logger.get_logger()
_REGEXES = {"BOOT_UP": r"<inf> app: Starting CHIP task",}
_BOOTUP_TIMEOUT = 10  # seconds
_RPC_TIMEOUT = 10  # seconds
# Button event dict: button id -> expected regex
# Button action details can be referred to
# https://github.com/project-chip/connectedhomeip/tree/master/examples/lighting-app/nrfconnect#device-ui
_BUTTON_EVENT_REGEXES = immutabledict.immutabledict({
    0: "Factory Reset Triggered",
    1: "Action has been completed",
    2: "New pairing for device",
    3: "NFC Tag emulation and BLE advertisement"})


class NRFPigweedLighting(nrf_connect_sdk_device.NRFConnectSDKDevice):
  """Device class for NRFPigweedLighting devices.

  This device class is for the Pigweed lighting application running
  on the nRF connect platform:
  https://github.com/project-chip/connectedhomeip/tree/master/examples/lighting-app/nrfconnect
  """
  DETECT_MATCH_CRITERIA = {
      detect_criteria.PigweedQuery.product_name: "j-link",
      detect_criteria.PigweedQuery.manufacturer_name: "segger",
      detect_criteria.PigweedQuery.app_type:
          pwrpc_utils.PigweedAppType.LIGHTING.value,
  }
  DEVICE_TYPE = "nrfpigweedlighting"
  _OWNER_EMAIL = "gdm-authors@google.com"
  _COMMUNICATION_KWARGS = {"protobufs": (button_service_pb2,
                                         lighting_service_pb2,
                                         device_service_pb2),
                           "baudrate": nrf_connect_sdk_device.BAUDRATE}

  def __init__(self,
               manager,
               device_config,
               log_file_name=None,
               log_directory=None):
    super().__init__(
        manager,
        device_config,
        log_file_name=log_file_name,
        log_directory=log_directory)
    self._regexes.update(_REGEXES)

  @decorators.DynamicProperty
  def firmware_version(self):
    """Firmware version of the device."""
    return self.pw_rpc_common.software_version

  @decorators.LogDecorator(logger)
  def reboot(self, no_wait: bool = False):
    """Reboots the device.

    Args:
      no_wait: Return before reboot completes.
    """
    self.pw_rpc_common.reboot(no_wait=no_wait,
                              rpc_timeout_s=_RPC_TIMEOUT,
                              bootup_logline_regex=self.regexes["BOOT_UP"],
                              bootup_timeout=_BOOTUP_TIMEOUT)

  @decorators.LogDecorator(logger)
  def factory_reset(self, no_wait: bool = False):
    """Factory resets the device.

    Args:
      no_wait: Return before reboot completes.
    """
    self.pw_rpc_common.factory_reset(
        no_wait=no_wait,
        rpc_timeout_s=_RPC_TIMEOUT,
        bootup_logline_regex=self.regexes["BOOT_UP"],
        bootup_timeout=_BOOTUP_TIMEOUT)

  @decorators.CapabilityDecorator(pwrpc_button_default.PwRPCButtonDefault)
  def pw_rpc_button(self):
    """PwRPCButtonDefault capability to send RPC command."""
    return self.lazy_init(
        pwrpc_button_default.PwRPCButtonDefault,
        device_name=self.name,
        expect_button_regexes=_BUTTON_EVENT_REGEXES,
        expect_timeout=_RPC_TIMEOUT,
        switchboard_call_expect=self.switchboard.call_and_expect)

  @decorators.CapabilityDecorator(pwrpc_common_default.PwRPCCommonDefault)
  def pw_rpc_common(self):
    """PwRPCCommonDefault capability to send RPC command."""
    return self.lazy_init(
        pwrpc_common_default.PwRPCCommonDefault,
        device_name=self.name,
        switchboard_call=self.switchboard.call,
        switchboard_call_expect=self.switchboard.call_and_expect)

  @decorators.CapabilityDecorator(pwrpc_light_default.PwRPCLightDefault)
  def pw_rpc_light(self):
    """PwRPCLight capability to send RPC command."""
    return self.lazy_init(pwrpc_light_default.PwRPCLightDefault,
                          device_name=self.name,
                          switchboard_call=self.switchboard.call)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Device controllers and capabilities built into GDM."""
import os.path
import shutil
from typing import Any, Dict

from gazoo_device import _version
from gazoo_device import config
from gazoo_device import data_types
from gazoo_device import detect_criteria
from gazoo_device import gdm_logger
from gazoo_device.auxiliary_devices import cambrionix
from gazoo_device.auxiliary_devices import dli_powerswitch
from gazoo_device.auxiliary_devices import efr32
from gazoo_device.auxiliary_devices import efr32_pigweed_lighting
from gazoo_device.auxiliary_devices import esp32
from gazoo_device.auxiliary_devices import esp32_pigweed_echo
from gazoo_device.auxiliary_devices import esp32_pigweed_locking
from gazoo_device.auxiliary_devices import nrf52840
from gazoo_device.auxiliary_devices import nrf_pigweed_lighting
from gazoo_device.auxiliary_devices import raspberry_pi
from gazoo_device.auxiliary_devices import unifi_poe_switch
from gazoo_device.auxiliary_devices import yepkit
from gazoo_device.capabilities import comm_power_default
from gazoo_device.capabilities import device_power_default
from gazoo_device.capabilities import event_parser_default
from gazoo_device.capabilities import fastboot_default
from gazoo_device.capabilities import file_transfer_adb
from gazoo_device.capabilities import file_transfer_docker
from gazoo_device.capabilities import file_transfer_echo
from gazoo_device.capabilities import file_transfer_scp
from gazoo_device.capabilities import flash_build_jlink
from gazoo_device.capabilities import package_management_android
from gazoo_device.capabilities import pwrpc_button_default
from gazoo_device.capabilities import pwrpc_common_default
from gazoo_device.capabilities import pwrpc_echo_default
from gazoo_device.capabilities import pwrpc_light_default
from gazoo_device.capabilities import pwrpc_lock_default
from gazoo_device.capabilities import shell_ssh
from gazoo_device.capabilities import switch_power_dli_powerswitch
from gazoo_device.capabilities import switch_power_ethernet
from gazoo_device.capabilities import switch_power_unifi_switch
from gazoo_device.capabilities import switch_power_usb_default
from gazoo_device.capabilities import switch_power_usb_with_charge
from gazoo_device.capabilities import usb_hub_default
from gazoo_device.capabilities.interfaces import comm_power_base
from gazoo_device.capabilities.interfaces import device_power_base
from gazoo_device.capabilities.interfaces import event_parser_base
from gazoo_device.capabilities.interfaces import fastboot_base
from gazoo_device.capabilities.interfaces import file_transfer_base
from gazoo_device.capabilities.interfaces import flash_build_base
from gazoo_device.capabilities.interfaces import package_management_base
from gazoo_device.capabilities.interfaces import pwrpc_button_base
from gazoo_device.capabilities.interfaces import pwrpc_common_base
from gazoo_device.capabilities.interfaces import pwrpc_echo_base
from gazoo_device.capabilities.interfaces import pwrpc_light_base
from gazoo_device.capabilities.interfaces import pwrpc_lock_base
from gazoo_device.capabilities.interfaces import shell_base
from gazoo_device.capabilities.interfaces import switch_power_base
from gazoo_device.capabilities.interfaces import switchboard_base
from gazoo_device.capabilities.interfaces import usb_hub_base
from gazoo_device.switchboard import communication_types
from gazoo_device.switchboard import switchboard

__version__ = _version.version
_PUBLIC_KEY_SUFFIX = ".pub"
logger = gdm_logger.get_logger()


def download_key(key_info: data_types.KeyInfo, local_key_path: str) -> None:
  """Raises an error with instructions on how to generate or obtain a key.

  Args:
    key_info: Information about key to download.
    local_key_path: File to which the key should be stored.

  Raises:
    RuntimeError: Key has to be retrieved or generated manually.
  """
  base_error = f"GDM doesn't come with built-in SSH key {key_info}. "

  if local_key_path.endswith(_PUBLIC_KEY_SUFFIX):
    private_key_path = local_key_path[:-len(_PUBLIC_KEY_SUFFIX)]
  else:
    private_key_path = local_key_path
  how_to_fix = (
      "Run 'ssh-keygen' to generate your own key. "
      f"Select {private_key_path} as the file to which the key should be "
      "saved. Leave the passphrase empty.")
  raise RuntimeError(base_error + how_to_fix)


def export_extensions() -> Dict[str, Any]:
  """Exports built-in device controllers, capabilities, communication types."""
  return {
      "primary_devices": [],
      "auxiliary_devices": [
          cambrionix.Cambrionix,
          dli_powerswitch.DliPowerSwitch,
          efr32.EFR32,
          efr32_pigweed_lighting.EFR32PigweedLighting,
          esp32.ESP32,
          esp32_pigweed_echo.ESP32PigweedEcho,
          esp32_pigweed_locking.ESP32PigweedLocking,
          nrf52840.NRF52840,
          nrf_pigweed_lighting.NRFPigweedLighting,
          raspberry_pi.RaspberryPi,
          unifi_poe_switch.UnifiPoeSwitch,
          yepkit.Yepkit,
      ],
      "virtual_devices": [],
      "communication_types": [
          communication_types.AdbComms,
          communication_types.DockerComms,
          communication_types.JlinkSerialComms,
          communication_types.PtyProcessComms,
          communication_types.SshComms,
          communication_types.SerialComms,
          communication_types.YepkitComms,
          communication_types.PigweedSerialComms,
      ],
      "detect_criteria":
          detect_criteria.DETECT_CRITERIA,
      "capability_interfaces": [
          comm_power_base.CommPowerBase,
          device_power_base.DevicePowerBase,
          event_parser_base.EventParserBase,
          fastboot_base.FastbootBase,
          file_transfer_base.FileTransferBase,
          flash_build_base.FlashBuildBase,
          package_management_base.PackageManagementBase,
          pwrpc_button_base.PwRPCButtonBase,
          pwrpc_common_base.PwRPCCommonBase,
          pwrpc_echo_base.PwRPCEchoBase,
          pwrpc_light_base.PwRPCLightBase,
          pwrpc_lock_base.PwRPCLockBase,
          shell_base.ShellBase,
          switchboard_base.SwitchboardBase,
          switch_power_base.SwitchPowerBase,
          usb_hub_base.UsbHubBase,
      ],
      "capability_flavors": [
          comm_power_default.CommPowerDefault,
          device_power_default.DevicePowerDefault,
          event_parser_default.EventParserDefault,
          fastboot_default.FastbootDefault,
          file_transfer_adb.FileTransferAdb,
          file_transfer_docker.FileTransferDocker,
          file_transfer_echo.FileTransferEcho,
          file_transfer_scp.FileTransferScp,
          flash_build_jlink.FlashBuildJLink,
          package_management_android.PackageManagementAndroid,
          pwrpc_button_default.PwRPCButtonDefault,
          pwrpc_common_default.PwRPCCommonDefault,
          pwrpc_echo_default.PwRPCEchoDefault,
          pwrpc_light_default.PwRPCLightDefault,
          pwrpc_lock_default.PwRPCLockDefault,
          shell_ssh.ShellSSH,
          switch_power_dli_powerswitch.SwitchPowerDliPowerswitch,
          switch_power_ethernet.SwitchPowerEthernet,
          switch_power_unifi_switch.SwitchPowerUnifiSwitch,
          switch_power_usb_default.SwitchPowerUsbDefault,
          switch_power_usb_with_charge.SwitchPowerUsbWithCharge,
          switchboard.SwitchboardDefault,
          usb_hub_default.UsbHubDefault,
      ],
      "keys": list(config.KEYS.values()),
  }



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Queries sent to devices during detection to determine their device type."""
import copy
import enum
import functools
import logging
import re
import subprocess
import typing
from typing import Any, Callable, Collection, Dict, List, Union

from gazoo_device import config
from gazoo_device import extensions
from gazoo_device.base_classes import auxiliary_device_base
from gazoo_device.base_classes import primary_device_base
from gazoo_device.capabilities.interfaces import switchboard_base
from gazoo_device.utility import host_utils
from gazoo_device.utility import http_utils
from gazoo_device.utility import pwrpc_utils
from gazoo_device.utility import usb_utils
import immutabledict
import requests

_DeviceClassType = Union[auxiliary_device_base.AuxiliaryDeviceBase,
                         primary_device_base.PrimaryDeviceBase]
_LOG_FORMAT = "<%(asctime)s> GDM-M: %(message)s"

_ADB_COMMANDS = immutabledict.immutabledict({})

_DOCKER_COMMANDS = immutabledict.immutabledict({
    "PRODUCT_NAME": "docker ps --filter id={} --format {{{{.Names}}}}",
})

_SSH_COMMANDS = immutabledict.immutabledict({
    "UNIFI_PRODUCT_NAME": "mca-cli-op info",
    "DLI_PRODUCT_NAME": "http://{address}/restapi/config/=brand_name/",
    "RPI_PRODUCT_NAME": "cat /proc/device-tree/model",
})

_UNIFI_MODEL_PREFIXES = ("USW-", "US-")


@functools.total_ordering
class QueryEnum(enum.Enum):
  """Allows comparison of enum properties for sorting purposes."""

  def __lt__(self, other):
    return self.name < other.name  # pylint: disable=comparison-with-callable


class AdbQuery(QueryEnum):
  pass


class DockerQuery(QueryEnum):
  product_name = "product_name"


class GenericQuery(QueryEnum):
  always_true = "always_true"


class PtyProcessQuery(QueryEnum):
  product_name = "product_name"


class SerialQuery(QueryEnum):
  product_name = "usb info product_name"


class SshQuery(QueryEnum):
  """Query names for detection for SshComms Devices."""
  is_dli = "is_dli_power_switch"
  is_rpi = "is_raspberry_pi"
  is_unifi = "is_unifi_switch"


class PigweedQuery(QueryEnum):
  """Query names for detection for PigweedSerialComms Devices."""
  app_type = "app_type"
  product_name = "usb info product_name"
  manufacturer_name = "usb info manufacturer_name"


def _docker_product_name_query(
    address: str,
    detect_logger: logging.Logger,
    create_switchboard_func: Callable[..., switchboard_base.SwitchboardBase]
) -> str:
  """Gets product name from docker device."""
  del create_switchboard_func  # Unused by _docker_product_name_query
  try:
    name = subprocess.check_output(
        _DOCKER_COMMANDS["PRODUCT_NAME"].format(address).split())
    name = name.decode()
  except subprocess.CalledProcessError as err:
    detect_logger.info("_docker_product_name_query failure: " + repr(err))
    return ""
  detect_logger.info("_docker_product_name_query response: {}".format(name))
  return name


def _always_true_query(
    address: str,
    detect_logger: logging.Logger,
    create_switchboard_func: Callable[..., switchboard_base.SwitchboardBase]
) -> bool:
  """Used when there is just one type of device for a communication type."""
  del address, create_switchboard_func  # Unused: query always returns True
  detect_logger.info("_always_true_query response: True")
  return True


def _is_dli_query(
    address: str,
    detect_logger: logging.Logger,
    create_switchboard_func: Callable[..., switchboard_base.SwitchboardBase]
) -> bool:
  """Determines if address belongs to dli power switch."""
  del create_switchboard_func  # Unused by _is_dli_query
  try:
    response = http_utils.send_http_get(
        _SSH_COMMANDS["DLI_PRODUCT_NAME"].format(address=address),
        auth=requests.auth.HTTPDigestAuth("admin", "1234"),
        headers={"Accept": "application/json"},
        valid_return_codes=[200, 206, 207],
        timeout=1)
    name = response.text
  except RuntimeError as err:
    detect_logger.info("_is_dli_query failure: " + repr(err))
    return False
  detect_logger.info("_is_dli_query response: {!r}".format(name))
  return "Power Switch" in name


def _is_rpi_query(
    address: str,
    detect_logger: logging.Logger,
    create_switchboard_func: Callable[..., switchboard_base.SwitchboardBase]
) -> bool:
  """Determines if address belongs to raspberry pi."""
  del create_switchboard_func  # Unused by _is_rpi_query
  try:
    name = host_utils.ssh_command(
        address,
        command=_SSH_COMMANDS["RPI_PRODUCT_NAME"],
        user="pi",
        key_info=config.KEYS["raspberrypi3_ssh_key"])
  except RuntimeError as err:
    detect_logger.info("_is_rpi_query failure: " + repr(err))
    return False
  detect_logger.info("_is_rpi_query response: {!r}".format(name))
  return "Raspberry Pi" in name


def _is_unifi_query(
    address: str,
    detect_logger: logging.Logger,
    create_switchboard_func: Callable[..., switchboard_base.SwitchboardBase]
) -> bool:
  """Determines if address belongs to unifi poe switch."""
  del create_switchboard_func  # Unused by _is_unifi_query
  try:
    mca_info = host_utils.ssh_command(
        address,
        _SSH_COMMANDS["UNIFI_PRODUCT_NAME"],
        user="admin",
        key_info=config.KEYS["unifi_switch_ssh_key"])
  except RuntimeError as err:
    detect_logger.info("_is_unifi_query failure: " + repr(err))
    return False
  detect_logger.info("_is_unifi_query response: {!r}".format(mca_info))

  for model_prefix in _UNIFI_MODEL_PREFIXES:
    if model_prefix in mca_info:
      return True
  return False


def _pty_process_name_query(
    address: str,
    detect_logger: logging.Logger,
    create_switchboard_func: Callable[..., switchboard_base.SwitchboardBase]
) -> str:
  """Returns product name from pty process comms address directory.

  Args:
    address: The communication address.
    detect_logger: The logger of device interactions.
    create_switchboard_func: Method to create the switchboard.
  """
  del create_switchboard_func  # Unused by _pty_process_name_query
  detect_logger.info("pty_process_name_query response: {}".format(address))
  return address


def usb_product_name_query(
    address: str,
    detect_logger: logging.Logger,
    create_switchboard_func: Callable[..., switchboard_base.SwitchboardBase]
) -> str:
  """Gets product name from usb_info."""
  del create_switchboard_func  # Unused by usb_product_name_query
  product_name = usb_utils.get_product_name_from_path(address).lower()
  detect_logger.info(
      "usb_product_name_query response: {}".format(product_name))
  return product_name


def _manufacturer_name_query(
    address: str,
    detect_logger: logging.Logger,
    create_switchboard_func: Callable[..., switchboard_base.SwitchboardBase]
) -> str:
  """Gets manufacturer name from usb_info."""
  del create_switchboard_func  # Unused by _manufacturer_name_query
  manufacturer = usb_utils.get_device_info(address).manufacturer.lower()
  detect_logger.info(
      "_manufacturer_name_query response: {}".format(manufacturer))
  return manufacturer


def _pigweed_application_query(
    address: str,
    detect_logger: logging.Logger,
    create_switchboard_func: Callable[..., switchboard_base.SwitchboardBase]
) -> str:
  """Gets Pigweed application type of the device.

  Args:
    address: The communication address.
    detect_logger: The logger of device interactions.
    create_switchboard_func: Method to create the switchboard.

  Returns:
    Pigweed application type.
  """
  file_handler = typing.cast(logging.FileHandler, detect_logger.handlers[0])
  log_path = file_handler.baseFilename
  app_type = pwrpc_utils.get_application_type(address,
                                              log_path,
                                              create_switchboard_func)
  detect_logger.info(f"_pigweed_application_query response {app_type}")
  return app_type


GENERIC_QUERY_DICT = immutabledict.immutabledict({
    GenericQuery.always_true: _always_true_query,
})

ADB_QUERY_DICT = immutabledict.immutabledict({
})

DOCKER_QUERY_DICT = immutabledict.immutabledict({
    DockerQuery.product_name: _docker_product_name_query,
})

PTY_PROCESS_QUERY_DICT = immutabledict.immutabledict({
    PtyProcessQuery.product_name: _pty_process_name_query,
})

SERIAL_QUERY_DICT = immutabledict.immutabledict({
    SerialQuery.product_name: usb_product_name_query,
})

SSH_QUERY_DICT = immutabledict.immutabledict({
    SshQuery.is_dli: _is_dli_query,
    SshQuery.is_rpi: _is_rpi_query,
    SshQuery.is_unifi: _is_unifi_query,
})

PIGWEED_QUERY_DICT = immutabledict.immutabledict({
    PigweedQuery.app_type: _pigweed_application_query,
    PigweedQuery.product_name: usb_product_name_query,
    PigweedQuery.manufacturer_name: _manufacturer_name_query,
})

DETECT_CRITERIA = immutabledict.immutabledict({
    "AdbComms": ADB_QUERY_DICT,
    "DockerComms": DOCKER_QUERY_DICT,
    "JlinkSerialComms": SERIAL_QUERY_DICT,
    "PtyProcessComms": PTY_PROCESS_QUERY_DICT,
    "SerialComms": SERIAL_QUERY_DICT,
    "SshComms": SSH_QUERY_DICT,
    "YepkitComms": GENERIC_QUERY_DICT,
    "PigweedSerialComms": PIGWEED_QUERY_DICT,
})


def determine_device_class(
    address: str,
    communication_type: str,
    log_file_path: str,
    create_switchboard_func: Callable[..., switchboard_base.SwitchboardBase]
) -> List[_DeviceClassType]:
  """Returns the device class(es) that matches the address' responses.

  Compares the device_classes DETECT_MATCH_CRITERIA to the device responses.

  Args:
    address: communication_address.
    communication_type: category of communication.
    log_file_path: local path to write log messages to.
    create_switchboard_func: Method to create the switchboard.

  Returns:
    list: classes where the device responses match the detect criteria.
  """
  detect_logger = _setup_logger(log_file_path)
  try:
    device_classes = get_communication_type_classes(communication_type)
    return find_matching_device_class(address, communication_type,
                                      detect_logger, create_switchboard_func,
                                      device_classes)
  finally:
    file_handler = detect_logger.handlers[0]
    file_handler.close()
    detect_logger.removeHandler(file_handler)


def find_matching_device_class(
    address: str,
    communication_type: str,
    detect_logger: logging.Logger,
    create_switchboard_func: Callable[..., switchboard_base.SwitchboardBase],
    device_classes: Collection[_DeviceClassType]) -> List[_DeviceClassType]:
  """Returns all classes where the device responses match the detect criteria.

  Args:
    address: communication_address.
    communication_type: category of communication.
    detect_logger: logs device interactions.
    create_switchboard_func: Method to create the switchboard.
    device_classes: device classes whose match criteria must be compared to.

  Returns:
    list: classes where the device responses match the detect criteria.
  """
  matching_classes = []
  responses = _get_detect_query_response(address, communication_type,
                                         detect_logger, create_switchboard_func)
  for device_class in device_classes:
    if _matches_criteria(responses, device_class.DETECT_MATCH_CRITERIA):
      matching_classes.append(device_class)
      detect_logger.info("{}: Match.".format(device_class.DEVICE_TYPE))
    else:
      detect_logger.info("{}: No Match.".format(device_class.DEVICE_TYPE))
  return matching_classes


def get_communication_type_classes(
    communication_type: str) -> List[_DeviceClassType]:
  """Returns classes with that communication type.

  Args:
    communication_type: category of communication.

  Returns:
    list: classes with that communication type.
  """
  all_classes = copy.copy(extensions.auxiliary_devices)
  all_classes += copy.copy(extensions.primary_devices)
  all_classes += copy.copy(extensions.virtual_devices)
  matching_classes = []
  for device_class in all_classes:
    if device_class.COMMUNICATION_TYPE == communication_type:
      matching_classes.append(device_class)
  return matching_classes


def _get_detect_query_response(
    address: str,
    communication_type: str,
    detect_logger: logging.Logger,
    create_switchboard_func: Callable[..., switchboard_base.SwitchboardBase]
) -> Dict[str, Any]:
  """Gathers device responses for all queries of that communication type.

  Args:
    address: communication_address
    communication_type: category of communication.
    detect_logger: logs device interactions.
    create_switchboard_func: Method to create the switchboard.

  Returns:
    Device responses keyed by query enum member.
  """
  query_responses = {}
  detect_queries = extensions.detect_criteria[communication_type]
  for query_name, query in detect_queries.items():
    try:
      query_responses[query_name] = query(
          address=address,
          detect_logger=detect_logger,
          create_switchboard_func=create_switchboard_func)
    except Exception as err:
      detect_logger.debug(
          f"failed getting detect query response for {address}: {err!r}")
      query_responses[query_name] = repr(err)

  return query_responses


def _matches_criteria(responses, match_criteria):
  """Checks if response dict matches match criteria.

  There are two categories of values in match_criteria: bool and regexp/str.
  Bools must match exactly, while regexp must find a match in the response
  value.

  Args:
    responses (dict): device responses keyed by query name.
    match_criteria (dict): match values keyed by query name.

  Returns:
    bool: whether or not responses meets match criteria
  """
  for entry, value in match_criteria.items():
    if isinstance(value, bool):
      if responses[entry] != value:
        return False
    else:
      if not re.search(value, responses[entry]):
        return False
  return True


def _setup_logger(log_file_path: str) -> logging.Logger:
  """Set up a logger to log device interactions to the detect file."""
  detect_logger = logging.getLogger(log_file_path)
  detect_logger.setLevel(logging.DEBUG)
  handler = logging.FileHandler(log_file_path)
  formatter = logging.Formatter(_LOG_FORMAT)
  handler.setFormatter(formatter)
  detect_logger.addHandler(handler)
  return detect_logger



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pseudoterminal transport class.

Some processes like the simulator shell script can fork/exec other
processes and need job control. A common way of overcoming this problem is
for the shell script to become a process leader. This is usually achieved by
using pseudoterminals.

More details: https://goo.gl/ckgF91

Pseudoterminals have a slightly different interface to reading and writing.

Example:
    os.read(self._process.stdout, size) in PTY
    self._process.stdout.read(size) in a regular process

This class will act as an adapter

"""
import os
import pty
import subprocess
import time

from gazoo_device import gdm_logger
from gazoo_device.switchboard import transport_properties
from gazoo_device.switchboard.transports import process_transport
import six

logger = gdm_logger.get_logger()


class PtyTransport(process_transport.ProcessTransport):
  """Perform process transport over pseudoterminal."""

  def __init__(self,
               comms_address,
               args="",
               auto_reopen=False,
               open_on_start=True):
    """Initialize the pty transport object with the given process properties.

    Args:
        comms_address (str): the process command and args to communicate
        args (str): additional args to pass to the main process to open
          communication.
        auto_reopen (bool): flag indicating transport should be reopened if
          unexpectedly closed.
        open_on_start (bool): flag indicating transport should be open on
          TransportProcess start.
    """
    self.comms_address = comms_address
    working_directory = os.path.dirname(comms_address)
    if working_directory:
      comms_address = comms_address.replace(working_directory, "./")
    else:
      working_directory = None
    super(PtyTransport, self).__init__(
        command=comms_address,
        args=args,
        auto_reopen=auto_reopen,
        open_on_start=open_on_start,
        working_directory=working_directory)
    self.primary = None
    self.secondary = None

  def close(self):
    """Closes the process if open."""
    if self.primary:
      os.close(self.primary)
      self.primary = None
    if self.secondary:
      os.close(self.secondary)
      self.secondary = None
    if hasattr(self, "_process") and self._process:
      if self._process.poll() is None:
        self._process.terminate()
        while self._process.poll() is None:
          time.sleep(0.001)
      self._process = None

  def _open(self):
    """Opens or reopens the process using the current property values.

    Raises:
        Exception: when invalid executable or options are provided.
    """
    if self._is_ready_to_open():
      try:
        # self.primary corresponds to the "master" file descriptor from pty.openpty().
        # self.secondary corresponds to the "slave" file descriptor from pty.openpty().
        self.primary, self.secondary = pty.openpty()
        self._process = subprocess.Popen(  # pylint: disable=subprocess-popen-preexec-fn
            self._args,
            stdin=self.secondary,
            stdout=self.secondary,
            stderr=self.secondary,
            cwd=self._cwd,
            close_fds=self._properties[transport_properties.CLOSE_FDS],
            preexec_fn=os.setsid  # This creates a new "process group"
        )
        # Set the stdin and stdout to self.primary
        self._process.stdout = self.primary
        self._process.stdin = self.primary
        os.close(self.secondary)
        self.secondary = None
      except Exception:
        self.close()
        raise

  def _read_data(self, size):
    try:
      return os.read(self._process.stdout, size)
    except OSError as err:
      logger.debug("Error in pty_process: {!r}".format(err))
      return b""

  def _write_data(self, data):
    if isinstance(data, six.integer_types):
      data = six.int2byte(data)
    os.write(self._process.stdin, data)
    return len(data)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Package management capability interface."""
import abc
from gazoo_device.capabilities.interfaces import capability_base


class PackageManagementBase(capability_base.CapabilityBase):
  """Abstract base class defining API for package management."""

  @abc.abstractmethod
  def install_package(self, package_path, *args, **kwargs):
    """Installs a package on the device.

    Args:
        package_path (str): the path to the package on host machine.
        *args (list): positional arguments.
        **kwargs (dict): optional keyword arguments.

    Raises:
        ValueError: when package_path is not valid.
        DeviceError: if failed to install the package.
    """

  @abc.abstractmethod
  def uninstall_package(self, package_name, *args, **kwargs):
    """Uninstalls a package on the device.

    Args:
        package_name (str): the name of the package to uninstall.
        *args (list): positional arguments.
        **kwargs (dict): optional keyword arguments.

    Raises:
        DeviceError: if failed to uninstall the package.
    """



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Internal Decorators used in device classes and capabilities.

Included are the following:

* Log Decorator
* CapabilityLogDecorator
* CapabilityDecorator
* health_check
* DynamicProperty
* PersistentProperty
* OptionalProperty


- BASIC USAGE:

***CapabilityDecorator***
Required for all capability properties in device classes.
Used to dynamically identify the capabilities and their flavors of a device
class.

@decorators.CapabilityDecorator(<capability_flavor>)

Example:

@decorators.CapabilityDecorator(file_transfer_scp.FileTransferScp)
    def file_transfer(self):
        return self.lazy_init(
            file_transfer_scp.FileTransferScp,
            ip_address_or_fn=self.ip_address,
            device_name=self.name,
            add_log_note_fn=self.switchboard.add_log_note,
            user=self._COMMUNICATION_KWARGS["username"],
            key_info=self._COMMUNICATION_KWARGS["key_info"])

If multiple flavors are declared, the flavor is not determined until initiation.
@decorators.CapabilityDecorator([file_transfer_adb.FileTransferAdb,
                                 file_transfer_scp.FileTransferScp])
    def file_transfer(self):
        if <condition>:
            return self.lazy_init(file_transfer_adb.FileTransferAdb,
                                  < ... keyword arguments ... >)
        else:
            return self.lazy_init(file_transfer_scp.FileTransferScp,
                                  < ... keyword arguments ... >)


***LogDecorator***
Required for all public methods in a device lcass that don't return a value.

This decorator will:
    1. Log a start message at the start of decorated method
    2. Log a success message at the end of decorated method if method succeeds
        2.1 The success message will also contain the method execution time.
    3. Log a skip message if method is skipped
        3.1 To skip a method, raise SkipExceptionError("YOUR_REASON.")
    4. Wrap all your exceptions in DeviceError (or a different wrap_type).
        4.1 The original exception type will be preserved in the error message.
    5. Add "<device> <method> failed. <Exception type>: <Exception message>."
        formatting to all exceptions.
        5.1 Note: don't specify the device name and method name in the exception
        message.
            This decorator does this for you!
            Example: use SomeError("Foo") instead of SomeError("Device <..>
            failed. Reason: Foo")

@decorators.LogDecorator(logger)
def your_method(self):
    pass

OR (to change the level)
Log_LEVEL can be one of NONE, DEBUG, INFO, WARNING, ERROR, CRITICAL
@decorators.LogDecorator(logger, level=decorators.<LOG_LEVEL>)
def your_method(self):
    pass


***CapabilityLogDecorator***
Required for all public methods in capability classes that don't return a value.
logger is the gdm_logger instance in the class

@decorators.CapabilityLogDecorator(logger)
def your_method(self):
    pass

OR (to change the level)
Log_LEVEL can be one of NONE, DEBUG, INFO, WARNING, ERROR, CRITICAL
@decorators.CapabilityLogDecorator(logger, level=decorators.<LOG_LEVEL>)
def your_method(self):
    pass

***health_check***
Used to identify all the public methods in a device class used for health
checks.
Note: The only public methods allowed in device classes are those identified in
the base template and health checks. All others must be in capabilities.

@decorators.health_check
def check_logs_streaming(self):

***DynamicProperty***
Used to identify dynamic properties, usually device states.

These properties change throughout testing. ex: firmware_version, pairing.state,
etc.

@decorators.DynamicProperty
def firmware_version(self):

***PersistentProperty***
Used to identify all persistent properties.

These are properties that are persistent throughout device lifetime.
ex: hardware_model, serial_number, etc. These are either populated during device
detection or static for a device class.

@decorators.PersistentProperty
def serial_number(self):

***OptionalProperty***
These are settable properties, usually initialized as None. They allow users to
enable additional capabilities or set properties undetectable via device
communication. For example, power_switch_port to enable power cycling via a
powerswitch or an alias.

@decorators.OptionalProperty
def alias(self):
"""
import functools
import inspect
import logging
import time
from typing import Any, Callable

from gazoo_device import errors
from gazoo_device import gdm_logger

logger_gdm = gdm_logger.get_logger()

# Enable specifying logger levels by decorators.<LEVEL> (so users don't have to import logging)
NONE = None
DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL

MESSAGES = {
    "START":
        "{device_name} starting {class_name}.{method_name}",
    "SKIP":
        "{device_name} {class_name}.{method_name} skipped. {skip_reason}.",
    "SUCCESS":
        "{device_name} {class_name}.{method_name} successful. It took "
        "{time_elapsed}s.",
    "FAILURE":
        "{device_name} {class_name}.{method_name} failed. {exc_name}: "
        "{exc_reason}"
}

DEFAULT_DEVICE_NAME = "Unknown_device"

# This attribute is added to decorated methods so it's possible
# to access the wrapped function. Same as "__wrapped__" attribute in Python 3.
WRAPS_ATTR_NAME = "_log_decorators_wraps"


class SkipExceptionError(Exception):
  """Used to tell info Error wrapper to skip last message."""


def unwrap(func):
  """Get the original function object of a wrapper function.

  Noop (returns func) if func is not decorated with decorator.

  Args:
      func (function): method, or property decorated with decorator

  Returns:
      function: the original function object (prior to decoration).
  """
  while hasattr(func, "__wrapped__"):
    func = func.__wrapped__
  return func


class CapabilityDecorator():
  """Decorator which defines capabilities in device classes.

  Usage example (from base_classes/raspbian_device.py):

  @decorators.CapabilityDecorator(file_transfer_scp.FileTransferScp)
  def file_transfer(self):
      return self.lazy_init(
          file_transfer_scp.FileTransferScp,
          ip_address_or_fn=self.ip_address,
          device_name=self.name,
          add_log_note_fn=self.switchboard.add_log_note,
          user=self._COMMUNICATION_KWARGS["username"],
          key_info=self._COMMUNICATION_KWARGS["key_info"])
  """

  def __init__(self, class_or_classes):
    """Initialize the capability decorator.

    Args:
        class_or_classes (object): single capability flavor class or list
          (or tuple) of capability flavor classes. GDM uses this to
          determine which capabilities are supported by a device class.

    Note: Input validation of class_or_classes is not done by the decorator
      at import time. Instead, it is performed by
      unit_tests/test_all_devices.py. This allows to avoid a circular
      import.
    """
    classes = class_or_classes
    if not isinstance(class_or_classes, (list, tuple)):
      classes = [classes]
    self._capability_classes = classes

  def __call__(self, fget):
    """Wrapper around CapabilityProperty to support decorator syntax.

    Args:
        fget (method): getter function (returns a capability instance). Note
          that most capability getter functions should use
          GazooDeviceBase.lazy_init().

    Returns:
        CapabilityProperty: property denoting a supported capability.
    """
    return CapabilityProperty(fget, classes=self._capability_classes)


class CapabilityProperty(property):
  """Property under which a capability is accessible in a device class.

  Do not use directly.

  Disambiguates device property definitions and device capability definitions.
  """
  fget: Callable[[Any], Any]  # Make pytype aware of the "fget" attribute.

  def __init__(self, fget, classes):
    """Initialize the capability property.

    Args:
        fget (method): getter function (returns a capability instance). Note
          that most capability getter functions should use
          GazooDeviceBase.lazy_init().
        classes (sequence): list or tuple of capability flavor classes. Used
          by GDM to determine which capabilities are supported by a device
          class.
    """
    super().__init__(fget)
    self.capability_classes = set(classes)


class LogDecorator():
  """Wraps GDM methods with standard logger messages (logged to the provided logger).

  Catches all exceptions and wraps them in DeviceError (or other wrap_type).
  Required for any public method that doesn't return a value.

  Note:
      GDM design asks for there to be a syntax for all info messsages and
      errors raised.
      This enforces that syntax.

      Example output when successful:
      cambrionix-1234 started factory_reset
      cambrionix-1234 successfully factory_reset in 10 s.

      Example out when failed:
      cambrionix-1234 started factory_reset
      cambrionix-1234 factory_reset failed. DeviceError: something failed.
  """

  def __init__(self,
               logger,
               level=INFO,
               wrap_type=errors.DeviceError,
               name_attr="name"):
    """Create a log decorator wrapper.

    Args:
        logger (logger): logger to use to print messages.
        level (int or None): logging level (see logging module level for
          integer values). (decorators.INFO, decorators.DEBUG,
          decorators.NONE) NONE disables logging messages (but errors are
          still wrapped).
        wrap_type (DeviceError): wrap the errors in this class if it's not
          an instance of it already. Valid values are DeviceError or its
          subclasses.
        name_attr (str): name of attribute containing the device name. For
          example, if device name is under self.device_name, the value
          should be "device_name".

    Raises:
        TypeError: if wrap_type is not a subclass of DeviceError.
    """
    if not issubclass(wrap_type, errors.DeviceError):
      raise TypeError(
          "Invalid wrap_type: {}. "
          "wrap_type can only be a subclass of DeviceError.".format(wrap_type))
    self.logger = logger
    self.level = level
    self.wrap_type = wrap_type
    self.name_attr = name_attr

  def __call__(self, func):
    """Wraps (decorates) the provided function.

    Args:
        func (function): function to be decorated

    Returns:
        function: a wrapper function (decorated input function)

    Raises:
        TypeError: incorrect type of func argument.
    """
    # Note: cannot use inspect.ismethod() here. At the time of decorator __init__ call,
    # python intepreter has not yet created the class of the method we're decorating.
    # At this time, the 'method' is still a function. Once all methods are created,
    # the python interpreter creates the class object, then binds the functions' self
    # arguments to make them 'methods.' Hence at this point func is considered a function
    # by python even if it's actually a method.
    if not callable(func):
      raise TypeError("Expected func to be callable, found {}.".format(func))

    error_template = (
        "Expected {} to be an instance method. "
        "Decorating {} methods is not supported; remove the decorator.")
    func_args = inspect.getfullargspec(func).args

    if func_args and func_args[0] == "cls":
      raise TypeError(error_template.format(func, "class"))
    if not func_args or func_args[0] != "self":
      raise TypeError(error_template.format(func, "static"))

    @functools.wraps(func)
    def wrapped_func(instance, *args, **kwargs):
      """Wraps (decorates) the given function.

      Args:
          instance (object): class instance
          *args (tuple): positional arguments to the wrapped function
          **kwargs (dict): keyword arguments to the wrapped function

      Raises:
          TracebackError: wrapped error (the traceback is added to __str__
          and __repr__).

      Note:
          this wrapper only works for methods, not functions (see the
          instance argument).

      Returns:
          object: same value as the wrapped function.
      """
      fmt_args = {
          "device_name": getattr(instance, self.name_attr, DEFAULT_DEVICE_NAME),
          "method_name": func.__name__,
          "class_name": self._find_defining_class_name(func, type(instance)),
          "skip_reason": None,
          "time_elapsed": None,
          "exc_name": None,
          "exc_reason": None
      }

      return_val = None
      method_skipped = False
      start_time = time.time()

      if self.level is not None:
        self.logger.log(self.level, MESSAGES["START"].format(**fmt_args))

      try:
        return_val = func(instance, *args, **kwargs)
      except SkipExceptionError as err:
        method_skipped = True
        fmt_args["skip_reason"] = str(err)
      except Exception as err:
        self._format_and_raise(fmt_args, err)

      if self.level is not None:
        if method_skipped:
          self.logger.log(self.level, MESSAGES["SKIP"].format(**fmt_args))
        else:
          fmt_args["time_elapsed"] = int(time.time() - start_time)
          self.logger.log(self.level, MESSAGES["SUCCESS"].format(**fmt_args))

      return return_val

    wrapped_func.__dict__[WRAPS_ATTR_NAME] = func
    return wrapped_func

  def _find_defining_class_name(self, method, current_class):
    """Finds the name of the class from which the method was inherited from.

    Args:
        method (func): method object.
        current_class (type): class to start the search from.

    Returns:
        str: name of the class defining the given method.
        None: defining class wasn't found in the class hierarchy.

    Note:
        There are 2 subtleties:
        * to correctly identify the class for super().<method> calls, the
        method check has to
          compare the method by identity (to avoid matching overridden
          methods).
        * methods may need to be unwrapped (from the log decorator) to make
        the identity
          comparison work.
        Doesn't work on class methods, but they are not decorated with the
        log decorator.
    """
    for a_class in current_class.__mro__:
      a_class_method = vars(a_class).get(method.__name__)
      if a_class_method and unwrap(a_class_method) is method:
        return a_class.__name__
    return None

  def _format_and_raise(self, fmt_args, err):
    """Format the exception message & type before re-raising."""
    wrapped_exc = err
    if (not isinstance(err, errors.CheckDeviceReadyError) and
        not isinstance(err, self.wrap_type)):
      fmt_args["exc_name"] = type(err).__name__
      fmt_args["exc_reason"] = str(err)
      # Wrap the error in a different type and reraise.
      reraise_msg = MESSAGES["FAILURE"].format(**fmt_args)
      wrapped_exc = self.wrap_type(reraise_msg)

    # Note: do nothing for CheckDeviceReadyErrors & subclasses.
    # These have two mandatory positional arguments in their signatures,
    # unlike the other errors, which typically have just one (the message).
    # It would be nice to get rid of the extra arguments (device name) in the signatures
    # and instead just let the log decorator add the device name to the error message.
    # However, refactoring these errors to use the log decorator is non-trivial.
    # This is because many of these are raised in plain functions (not methods),
    # which are not supported by this log decorator...
    raise wrapped_exc


class CapabilityLogDecorator(LogDecorator):
  """Adds standardized log messages to capability methods."""

  def __init__(self,
               logger,
               level=INFO,
               wrap_type=errors.DeviceError,
               name_attr="_device_name"):
    """Log decorator for capabilities (with a different name_attr default value).

    Args:
        logger (logger): logger to use to print messages.
        level (int or None): logging level (see logging module level for
          integer values). For example, level=INFO or level=NONE. NONE
          disables logging messages (but errors are still wrapped).
        wrap_type (DeviceError): wrap the errors in this class if it's not
          an instance of it already. Valid values are DeviceError or its
          subclasses.
        name_attr (str): name of attribute containing the device name. For
          example, if device name is under self.device_name, the value
          should be "device_name".

    Note: All capabilities need to have a device name reference in the
      capability class. This is required for log and error messages in
      multi-device tests.
    """
    super().__init__(
        logger, level=level, wrap_type=wrap_type, name_attr=name_attr)


decorators_factory = LogDecorator(logger_gdm, level=DEBUG)


def health_check(func):
  """Health Check Decorator.  Returns silenced log decorator."""
  func.__health_check__ = True
  return decorators_factory(func)


class DynamicProperty(property):
  """A property that is dynamic and involves a device query to return.

  These properties may be settable if there is a corresponding setter property
  function.
  """
  # Make pytype aware of the "setter" attribute.
  setter: Callable[[Callable[[Any, Any], None]], None]

  def __init__(self, fget, fset=None, fdel=None, doc=None):
    if not doc:
      doc = fget.__doc__
    super().__init__(fget, fset=fset, fdel=fdel, doc=doc)
    self.name = fget.__name__


class OptionalProperty(property):
  """A property that can be set and usually defaults to null.

  The property that is read from device.props["optional"] and exists in
  device_options.json
  The property is settable without specifying an explicity setter - the value
  is modified directly in device_options.json.

  These values may be reset if a device is deleted and then re-added.
  If the device is redetected, optional properties will be preserved.
  """
  # Make pytype aware of the "setter" attribute.
  setter: Callable[[Callable[[Any, Any], None]], None]

  def __init__(self, fget, fset=None, fdel=None, doc=None):
    if not doc:
      doc = fget.__doc__
    super().__init__(fget, fset=fset, fdel=fdel, doc=doc)
    self.name = fget.__name__


class PersistentProperty(property):
  """A property that is persistent throughout device interaction and set on device detection.

  Persistent properties exist in devices.json file.
  Is static across device class or can only be reset during device detection.
  """

  def __init__(self, fget):
    super().__init__(fget, doc=fget.__doc__)
    self.name = fget.__name__



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""USB hub capability interface.

Abstract Base class for the usb_hub capability.

The usb_hub capability is intended to be used by primary devices that require
the ability to
get or change the USB power mode for a configured port on a USB hub.

The configured USB hub must support the switch_power capability.
"""
import abc
from gazoo_device.capabilities.interfaces import capability_base


class UsbHubBase(capability_base.CapabilityBase):
  """Abstract base class for USB hub capability."""

  @abc.abstractproperty
  def name(self):
    """The name of the usb hub.

    Returns:
        str: usb hub name.
    """

  @abc.abstractproperty
  def device_port(self):
    """The usb hub port number used by device.

    Returns:
        int: port number on usb hub.
    """

  @abc.abstractproperty
  def supported_modes(self):
    """Get the USB power modes supported by the USB hub."""

  @abc.abstractmethod
  def get_device_power(self):
    """Gets usb port mode if set.

    Returns:
        str: 'sync', 'charge', or 'off'

    Raises:
        DeviceError: if usb_hub and usb_port are not properly set.
    """

  @abc.abstractmethod
  def set_device_power(self, mode):
    """Turns associated powered usb hub port, if available, power state to sync, off, charge.

    Args:
        mode (str): power mode to set USB hub port to ("sync", "off",
          "charge")

    Raises:
        DeviceError: if invalid mode provided or usb_hub management is not a
        valid option

    Notes:
        'sync' is data and power on, 'charge' is power only on, 'off' is both
        off.
    """

  @abc.abstractmethod
  def power_off(self, port=None):
    """This command powers off the port specified or all ports if port is None.

    Args:
        port (int): identifies which hub port to power off

    Raises:
         DeviceError: port number invalid.
    """

  @abc.abstractmethod
  def power_on(self, port=None, data_sync=True):
    """This command powers on the port specified or all ports if port is None.

    Args:
        port (int): identifying which hub port to power on
        data_sync (bool): True if data should be enabled, false for power only

    Raises:
         DeviceError: port number invalid.
    """
